#!/bin/sh
# the next line restarts using wish \
exec wish "$0" "$@"


#############################################################################
# PgcEdit.tcl Copyright 2004 r0lZ
# Program main entry.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

#############################################################################
#
# PgcEdit requires Tcl/Tk 8.4 or greather
#
#############################################################################

set ::program_version 0.1.2


wm withdraw .
console hide	;# necessary for freeWrap 5.61
update idletasks

if {$tcl_version < 8.4} {
	tk_messageBox -icon error -title "PgcEdit $::program_version error" -type ok -parent . \
		-message "PgcEdit requires Tcl/Tk version 8.4 or greather.\nYou are currently running Tcl/Tk version $tcl_patchLevel.\nSorry."
	exit 1000
}

#############################################################################
## Debug print.
## Usage: set ::debug <triggerlevel>
##        dprint <level> <message>

proc dprint {l t} {
	catch {if {$::debug>=$l} {puts $t}}
}

#############################################################################
## Procedure:  init_user_prefs
## Init user modifiables constants
proc init_user_prefs {} {
	global argv0

	set ::config(restore_defaults) false

	if {$::tcl_platform(platform) == "windows"} {
		set rootdir "C:"
		foreach i {C: D: E: F: G: H: I: J: K: L: M: N: O: P: Q: R: S: T: U: V: W: X: Y: Z:} {
			if [file exists $i] {
				set rootdir "$i\\"
			}
		}
		set progfiles "${rootdir}Program files\\"
	} else {
		set rootdir {/}
		set progfiles {/}
	}

	# Automatically save setups on program exit?
	set ::config(save_setups_on_exit) true

	# Global font
	label .testfont
	set ::config(global_font) [.testfont cget -font]
	destroy .testfont

	# Fixed-size font
	set ::config(fixed_width_font) {{Courier} 9}

	# Tearoff menus
	set ::config(tearoff_menus) true

	# options set automatically by the program:

	# Automatically launch Open DVD dialog at startup?
	set ::config(auto_open_new_dvd) false

	# Automatically create backup when a DVD is opened?
	set ::config(auto_create_backup) true

	# Automatically open the selector at startup?
	set ::config(selector) true

	# Automatically show the action buttons at startup?
	set ::config(buttons) true

	# Automatically adjust Command Editor window size?
	set ::config(adjust_win_size) false

	# The startup directory for the DVD file browser:
	set ::config(initialdir) $rootdir

	# The startup directory for the Load Menu file browser:
	set ::config(importmenuinitialdir) $rootdir

	# The startup directory for the scratchbook file browser:
	set ::config(scratchbook_initialfile) $rootdir

	# The startup filename for the dump file browser:
	set ::config(dumpinitialfile) "${rootdir}dvd_dump.txt"

	# Size and position of the main window
	set ::config(main_win_geometry) 700x400

	# Position of the Command Editor window
	set ::config(ced_win_geometry) ""

	# Position of the Command Editor window
	set ::config(selector_win_geometry) ""

	# Debug level
	set ::config(debug) 0

	# Overwrite default values with user defined ones
	set configfile [file join $::env(HOME) ".pgcedit.cfg"]
	if [file exists $configfile] {
		if {[catch {source $configfile} err] != 0} {
			tk_messageBox -icon error -title "PgcEdit $::program_version error" -type ok \
				-message "Error in config file $configfile:\n$err" -parent .
		}
	}

	# if ::config(initialdir) does not exists, find the first directory that exists
	while {![file exists $::config(initialdir)]} {
		set newdir [file dirname $::config(initialdir)]
		if {$::config(initialdir) == $newdir} {
			set ::config(initialdir) $rootdir
		} else {
			set ::config(initialdir) $newdir
		}
	}

	# if ::config(importmenuinitialdir) does not exists, find the first directory that exists
	while {![file exists $::config(importmenuinitialdir)]} {
		set newdir [file dirname $::config(importmenuinitialdir)]
		if {$::config(importmenuinitialdir) == $newdir} {
			set ::config(importmenuinitialdir) $rootdir
		} else {
			set ::config(importmenuinitialdir) $newdir
		}
	}

	# if ::config(dumpinitialfile) does not exists, find the first directory that exists
	set pathname [file dirname $::config(dumpinitialfile)]
	set filename [file tail    $::config(dumpinitialfile)]
	while {![file exists $pathname]} {
		set newdir [file dirname $pathname]
		if {$pathname == $newdir} {
			set pathname $rootdir
		} else {
			set pathname $newdir
		}
	}
	set ::config(dumpinitialfile) [file join $pathname $filename]

	# Tools
	set ::config(tools) {}

	# Overwrite default values with user defined ones
	set configfile [file join $::env(HOME) ".pgcedit.tools"]
	if [file exists $configfile] {
		if {[catch {source $configfile} err] != 0} {
			tk_messageBox -icon error -title "PgcEdit $::program_version error" -type ok \
				-message "Error in config file $configfile:\n$err" -parent .
		}
	}

	# set the "associate PgcEdit with .XXX" menu flags
	# based upon current registry associations
	# Only used if PgcEdit is freeWrapped and run under Windows
	if {$::tcl_platform(platform) == "windows" && [info exists ::freewrap::progname]} {
		set ::config(shortcut_startmenu) \
				[::winshellext::is_shortcut_to_PgcEdit startmenu]
		set ::config(shortcut_sendto) \
				[::winshellext::is_shortcut_to_PgcEdit sendto]
		set ::config(shortcut_desktop) \
				[::winshellext::is_shortcut_to_PgcEdit desktop]
		set ::config(assoicated_with_IFO) \
				[::winshellext::shell_is_associated ".IFO" "PgcEdit.exe"]
		set ::config(assoicated_with_BUP) \
				[::winshellext::shell_is_associated ".BUP" "PgcEdit.exe"]
		set ::config(assoicated_with_VOB) \
				[::winshellext::shell_is_associated ".VOB" "PgcEdit.exe"]
	}
}

#############################################################################
## Procedure:  save_user_prefs
## save setups
proc {save_user_prefs} {} {
	set txt "# PgcEdit config file.  Saved automatically.  Do not edit!\n"
	set txt "${txt}\nset ::config(global_font) \{$::config(global_font)\}"
	set txt "${txt}\nset ::config(fixed_width_font) \{$::config(fixed_width_font)\}"
	set txt "${txt}\nset ::config(tearoff_menus) $::config(tearoff_menus)"
	set txt "${txt}\nset ::config(save_setups_on_exit) $::config(save_setups_on_exit)"
	set txt "${txt}\nset ::config(auto_open_new_dvd) $::config(auto_open_new_dvd)"
	set txt "${txt}\nset ::config(auto_create_backup) $::config(auto_create_backup)"
	set txt "${txt}\nset ::config(selector) $::config(selector)"
	set txt "${txt}\nset ::config(buttons) $::config(buttons)"
	set txt "${txt}\nset ::config(adjust_win_size) $::config(adjust_win_size)"
	set txt "${txt}\nset ::config(initialdir) \{[string map {\\ \\\\} [file nativename $::config(initialdir)]]\}"
	set txt "${txt}\nset ::config(importmenuinitialdir) \{[string map {\\ \\\\} [file nativename $::config(importmenuinitialdir)]]\}"
	set txt "${txt}\nset ::config(scratchbook_initialfile) \{[string map {\\ \\\\} [file nativename $::config(scratchbook_initialfile)]]\}"
	set txt "${txt}\nset ::config(dumpinitialfile) \{[string map {\\ \\\\} [file nativename $::config(dumpinitialfile)]]\}"
	set txt "${txt}\nset ::config(main_win_geometry) [winfo geometry .]"
	set txt "${txt}\nset ::config(ced_win_geometry) [winfo geometry .cmdeditor]"
	if [winfo exists .selector] {
		set txt "${txt}\nset ::config(selector_win_geometry) [winfo geometry .selector]"
	} else {
		set txt "${txt}\nset ::config(selector_win_geometry) $::config(selector_win_geometry)"
	}
	if {$::config(debug) > 0} {
		set txt "${txt}\nset ::config(debug) $::config(debug)"
	}
	set txt "${txt}"

	saveFile [file join $::env(HOME) ".pgcedit.cfg"] auto $txt
}

#############################################################################
## Procedure:  cleanexit
## save setups and exit
proc {cleanexit} {code} {
	if {$::dvd_modified && $code == 0} {
		set r [tk_messageBox -message "The DVD has been modified.\n\nOK to quit?" \
				-icon warning -title "PgcEdit" -type yesno -default no]
		if {$r == "no"} return
	}
	#bind . <Destroy> {}
	if {$::config(save_setups_on_exit) && $::config(restore_defaults)==false} {
		save_user_prefs
	}
	if $::config(debug) {
		tk_messageBox -icon info -message "Exiting (code=$code)" -title "PgcEdit" -type ok
	}
	exit $code
}

#############################################################################
## Procedure:  errmsg
## Display an error message
proc {errmsg} {errnum msg1 msg2} {
	# Display a message
	puts "Error $errnum: $msg1"
	puts "$msg2"
	if $::config(debug) {
		console show
	}
	tk_messageBox -icon error -title "PgcEdit $::program_version error" -type ok \
		-message "Error $errnum $msg1\n$msg2" -parent .
}

#############################################################################
## Procedure:  readFile
## read a file and return the datas
## return "" in case of error
proc readFile {filename translation} {
	set errnum [catch {set in [open $filename r]} err]
	if {$errnum} {
		errmsg $errnum "opening file for read" $err
		return ""
	}
	fconfigure $in -translation $translation
	set errnum [catch {set data [read $in]} err]
	if {$errnum} {
		close $in
		errmsg $errnum "reading file" $err
		return ""
	}
	close $in
	return $data
}

#############################################################################
## Procedure:  saveFile
## save datas in a file
## return error number, or 0 for OK
proc saveFile {filename translation data} {
	set errnum [catch {set out [open $filename w]} err]
	if {$errnum} {
		errmsg $errnum "opening file for write" $err
		return $errnum
	}
	fconfigure $out -translation $translation
	if {$translation == "binary"} {
		set errnum [catch {puts -nonewline $out $data} err]
	} else {
		set errnum [catch {puts $out $data} err]
	}
	if {$errnum} {
		close $out
		errmsg $errnum "writing file" $err
		return $errnum
	}
	close $out
	return $errnum
}

#############################################################################
## Procedure:  b2x
## Convert a binary string to hexadecimal
proc {b2x} {binarystring} {
	binary scan $binarystring H* hex
	return $hex
}

#############################################################################
## Procedure:  b2i
## Convert a binary string to decimal
proc {b2i} {binarystring} {
	binary scan $binarystring H* hex
	return [expr 0x$hex]
}

#############################################################################
## PROGRAM MAIN ENTRY
#############################################################################

# Parse arguments
if {$argc == 1 && ([lindex $argv 0] == "/?" || [lindex $argv 0] == "-h")} {
	puts "PgcEdit v $::program_version"
	puts "Command line usage:"
	puts "> PgcEdit \[\"DVD pathname\"\]"
	puts "or"
	puts "> PgcEdit \[\"DVD pathname\any file\"\]"
	puts "The second form is used to associate PgcEdit with IFO (or BUP) files."
	exit 0
}

# Full path is needed for FreeWrap
if {[info exists ::freewrap::progname]} {
	set config(apppath) "/Tcl/work/PGCEDIT"
	if {$::tcl_platform(platform) == "windows"} {
		source $config(apppath)/lib/winshellext.tcl
	}
} else {
	set config(apppath) "[file dirname $argv0]"
}

# init user preferences
init_user_prefs

# disable the dprint routine if it is not needed, for speed purposes
if {$::config(debug) == 0} {
	proc dprint {args} {}
}

# init some system defaults
# we need to know some widget default colors
set config(systembuttonface) [[button .w] cget -background]; destroy .w
if {$::tcl_platform(platform) == "windows"} {
	set config(systemhighlight) systemhighlight
} else {
	set config(systemhighlight) gray75
}
option add *font $::config(global_font)
option add *tearOff $::config(tearoff_menus)
option add *Scale*troughColor white
option add *Scale*sliderLength 15
option add *Scale*highlightThickness 1
option add *Scale*borderWidth 1
option add *Scale*orient horizontal
option add *Menubutton*relief raised
option add *Menubutton*indicatorOn true
option add *Menubutton*highlightThickness 1
option add *Menubutton*takeFocus _focus_menubutton
proc _focus_menubutton {w} {if {[$w cget -state]=="disabled"} {return 0} else {return 1}}

# read program modules
source $config(apppath)/lib/tk_menu_bugfix.tcl	;# MenuUnpost Tk bug fixed by r0lZ
source $config(apppath)/lib/readdvd.tcl
source $config(apppath)/lib/writedvd.tcl
source $config(apppath)/lib/disasm.tcl
source $config(apppath)/lib/utils.tcl
source $config(apppath)/lib/toolscfg.tcl
source $config(apppath)/lib/main.tcl			;# main program starts here
source $config(apppath)/lib/cmdedit.tcl
source $config(apppath)/lib/pgceditor.tcl
source $config(apppath)/lib/srptinfo.tcl
source $config(apppath)/lib/streams.tcl
source $config(apppath)/lib/stringreq.tcl
source $config(apppath)/lib/fontreq.tcl
source $config(apppath)/lib/calculator.tcl

# initialize the command editor
::ced::cmdeditor
