#############################################################################
# main.tcl Copyright 2004 r0lZ
# PgcEdit main window GUI and procedures.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::main {

#############################################################################
## Procedure:  about
## Display the about message
proc about {} {
	set txt "Version $::program_version"
	set txt "$txt\nCopyright r0lZ, 2004"
	set txt "$txt\n\nPgcEdit is free software"
	set txt "$txt\ndistributed under the GPL license"
	set txt "$txt\n(GNU General Public License)"
	set txt "$txt\n\nUse PgcEdit at your own risk!"
	set txt "$txt\n\nBased on DVD Informations by MPUCoder:"
	set txt "$txt\nhttp://www.mpucoder.com/dvd/"
	set txt "$txt\n\nUsing Tcl/Tk version $::tcl_patchLevel"
	if {[info exists ::freewrap::progname]} {
		set txt "$txt\nWrapped with FreeWrap $::freewrap::patchLevel"
	}
	set txt "$txt\n\nPgcEdit home:"
	set txt "$txt\nhttp://home.tiscali.be/debie.roland/pgcedit/"

	toplevel .about
	wm withdraw .about
	wm title .about "About PgcEdit"
	wm transient .about .
	wm overrideredirect .about 1
	wm resizable .about 0 0

	frame .about.a -borderwidth 1 -relief sunken
	pack  .about.a -side top -fill both -expand 1
	frame .about.a.b -borderwidth 6 -relief ridge
	pack  .about.a.b -side top -fill both -expand 1 -anchor center
	frame .about.a.b.pad1 -height 20
	pack  .about.a.b.pad1 -side top -fill y -expand 1
	label .about.a.b.lab1 -text "PgcEdit" -bd 2 -relief sunken \
			-padx 12 -pady 10 -background #EEEEEE
	pack  .about.a.b.lab1 -side top -padx 20 -anchor center -pady 8
	.about.a.b.lab1 configure -font "\"[lindex [.about.a.b.lab1 cget -font] 0]\" 14"
	label .about.a.b.lab2 -text $txt
	pack  .about.a.b.lab2 -side top -padx 20 -anchor center
	frame .about.a.b.pad2 -height 20
	pack  .about.a.b.pad2 -side top -fill y -expand 1

	update
	if {[winfo reqwidth .about] > [winfo reqheight .about]} {
		set size [winfo reqwidth .about]
	} else {
		set size [winfo reqheight .about]
	}
	wm geometry .about ${size}x$size
	update
	#set wx [expr ([winfo screenwidth  .about]-[winfo reqwidth  .about])/2]
	#set wy [expr ([winfo screenheight .about]-[winfo reqheight .about])/2]
	set wx [expr [winfo rootx .]+(([winfo width  .]-[winfo reqwidth  .about])/2)]
	set wy [expr [winfo rooty .]+(([winfo height .]-[winfo reqheight .about])/2)-25]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry .about +$wx+$wy

	wm deiconify .about
	grab set .about
	focus .about
	bind .about <ButtonRelease-1> {destroy .about}
	bind .about <Escape> {destroy .about}
}

proc showtextfile {w title file} {
	set txt "File \"$file\" not found!"
	catch {set txt [string map {"\014" ""} [readFile $file auto]]}
	showtext $w $title $txt
}

proc showtext {w title txt} {
	set winname [string trim $w "."]
	set searchstringvar ::main::${winname}_searchstring
	set searchcasevar   ::main::${winname}_searchcase
	set searchregexpvar ::main::${winname}_searchregexp

	if {! [winfo exists $w]} {
		if {! [info exists $searchregexpvar]} {
			set $searchstringvar ""
			set $searchcasevar 0
			set $searchregexpvar "-exact"
		}

		toplevel $w
		wm withdraw $w
		wm title $w $title
		bind $w <Escape> "destroy $w"

		frame $w.a
		pack  $w.a -side top -fill both -expand 1 -padx 4 -pady 4
		text  $w.a.txt -xscrollcommand "$w.a.xs set" \
				-yscrollcommand "$w.a.ys set" \
				-font $::config(fixed_width_font) -wrap none
		scrollbar $w.a.ys -command "$w.a.txt yview"
		scrollbar $w.a.xs -command "$w.a.txt xview" -orient horizontal
		grid  $w.a.txt -column 0 -row 0 -sticky nesw
		grid  $w.a.ys  -column 1 -row 0 -sticky ns
		grid  $w.a.xs  -column 0 -row 1 -sticky ew
		grid columnconf $w.a 0 -weight 1
		grid rowconf    $w.a 0 -weight 1

		frame $w.b
		pack  $w.b -side top -pady 2 -fill x
		frame  $w.b.s -bd 2 -relief groove
		pack   $w.b.s -side left -padx 4 -fill x -expand 1
		label  $w.b.s.l -text "Find"
		pack   $w.b.s.l -side left
		entry  $w.b.s.e -textvariable $searchstringvar
		bind   $w.b.s.e <Return> "::main::searchtext $w -forwards"
		pack   $w.b.s.e -side left -fill x -padx 2 -expand 1
		button $w.b.s.b1 -text "<" -width 2 -command "::main::searchtext $w -backwards"
		pack   $w.b.s.b1 -side left -padx 2
		button $w.b.s.b2 -text ">" -width 2 -command "::main::searchtext $w -forwards"
		pack   $w.b.s.b2 -side left -padx 2
		frame  $w.b.s.cf
		pack   $w.b.s.cf -side left
		checkbutton $w.b.s.cf.c1 -bd 0 -text "match case" \
				-variable $searchcasevar
		pack   $w.b.s.cf.c1 -side top -anchor w
		checkbutton $w.b.s.cf.c2 -bd 0 -text "regular expression" \
				-variable $searchregexpvar \
				-onvalue "-regexp" -offvalue "-exact"
		pack   $w.b.s.cf.c2 -side top -anchor w

		button $w.b.save -text Save -width 8 -command "::main::savetext $w"
		pack   $w.b.save -side left -padx 4
		button $w.b.ok -text Close -width 8 -command "destroy $w"
		pack   $w.b.ok -side left -padx 4

		# window geometry
		update
		set wx [expr [winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)]
		set wy [expr [winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25]
		if {$wx < 10} {set wx 10}
		if {$wy < 10} {set wy 10}
		wm geometry $w +$wx+$wy
	}

	$w.a.txt delete 1.0 end
	$w.a.txt insert 1.0 $txt
	$w.a.txt mark set insert 1.0

	wm deiconify $w
}
proc savetext {w} {
	# Get input filename interactively
	set initialpath [file dirname $::config(dumpinitialfile)]
	set filename [tk_getSaveFile -filetypes {{"Text files" {*.txt}}} \
		-initialdir $initialpath -parent $w \
		-title "PgcEdit $::program_version: save text to file..."]
	if {$filename == ""} {return}
	set ::config(dumpinitialfile) $filename

	saveFile $filename auto [$w.a.txt get 1.0 end]
}
proc searchtext {w direction} {
	set winname [string trim $w "."]
	set searchstringvar ::main::${winname}_searchstring
	set searchcasevar   ::main::${winname}_searchcase
	set searchregexpvar ::main::${winname}_searchregexp

	set idx [$w.a.txt index insert]
	if {$direction == "-forwards"} {
		set idx "$idx + 1 chars"
	} else {
		set idx "$idx - 1 chars"
	}
	if [subst $$searchcasevar] {
		set err [catch {set found [$w.a.txt search $direction [subst $$searchregexpvar] \
				-- [subst $$searchstringvar] $idx]} errstr]
	} else {
		set err [catch {set found [$w.a.txt search $direction [subst $$searchregexpvar] \
				-nocase -- [subst $$searchstringvar] $idx]} errstr]
	}
	if {$err != 0} {
		tk_messageBox -title "Search error" -icon error \
				-message [string map {": " ":\n"} $errstr] -parent $w -type ok
	} else {
		if {$found != ""} {
			$w.a.txt mark set insert $found
			$w.a.txt see $found
			focus $w.a.txt
		} else {
			bell
		}
	}
}

#############################################################################
## Procedure:  Create_Backup
## If the backup directory already exists, then ask the user for overwrite, or
## if the ifnotexists arg is true, don't ask and don't overwrite
proc Create_Backup {ifnotexists} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set backupdir "$::dvddir/PgcEdit_backup"
	if [file exists $backupdir] {
		if $ifnotexists {
			return
		} else {
			set r [tk_messageBox -title "PgcEdit: Backup" -icon question \
				-message "Backup already exists!\nOverwrite?" -parent . \
				-type okcancel -default cancel]
			if {$r == "cancel"} return
			file delete -force -- "$backupdir/VIDEO_TS.IFO"
			for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
				file delete -force -- "$backupdir/VTS_[format %02i $vts]_0.IFO"
			}
			file delete -force -- $backupdir
			set dummy [file exists $backupdir]	;# wait till directory is deleted
		}
	}
	set err [catch {file mkdir $backupdir} errtxt]
	if {$err > 0} {
		errmsg $err "Error creating backup" $errtxt
	} else {
		file copy -force -- "$::dvddir/VIDEO_TS.IFO" $backupdir
		for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
			file copy -force -- "$::dvddir/VTS_[format %02i $vts]_0.IFO" $backupdir
		}
	}
}

#############################################################################
## Procedure:  Restore_Backup
proc Restore_Backup {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set backupdir [file join $::dvddir PgcEdit_backup]
	if {! [file exists $backupdir]} {
		tk_messageBox -title "PgcEdit: Backup" -icon error \
			-message "Backup not found!" -parent . -type ok
		return
	}
	file copy -force -- "$backupdir/VIDEO_TS.IFO" $::dvddir
	file copy -force -- "$backupdir/VIDEO_TS.IFO" "$::dvddir/VIDEO_TS.BUP"
	for {set vts 1} {$vts <= 99} {incr vts} {
		if [file exists "$backupdir/VTS_[format %02i $vts]_0.IFO"] {
			file copy -force -- "$backupdir/VTS_[format %02i $vts]_0.IFO" $::dvddir
			file copy -force -- "$backupdir/VTS_[format %02i $vts]_0.IFO" \
					"$::dvddir/VTS_[format %02i $vts]_0.BUP"
		}
	}
	set r [tk_messageBox -title "PgcEdit: Backup" -icon question \
		-message "Backup restored.\nReload DVD in editor?" -parent . \
		-type okcancel -default ok]
	if {$r == "ok"} ReRead_DVD
}

#############################################################################
## Procedure:  Delete_Backup
proc Delete_Backup {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set backupdir "$::dvddir/PgcEdit_backup"
	if {! [file exists $backupdir]} {
		tk_messageBox -title "PgcEdit: Backup" -icon error \
			-message "Backup not found!" -parent . -type ok
		return
	}
	file delete -force -- $backupdir
}

#############################################################################
## Procedure:  Write_DVD
## write the current DVD to disk.
## Get output filename interactively if saveas is true, else use global $::dvddir
proc Write_DVD {saveas} {

	if {! [info exists ::dvddir]} {
		bell
		return
	}

	if $saveas {
		set destdir [tk_chooseDirectory -initialdir $::dvddir \
			-title "PgcEdit $::program_version: Save IFOs as:"]
		if {$destdir == ""} {return -1}
		if {! [file exists $destdir]} {
			set r [tk_messageBox -title "PgcEdit: Save IFOs as" -icon question \
					-message "Folder \"$destdir\" does not exists.\nCreate ?" \
					-parent . -type yesno]
			if {$r == "no"} {
				return -1
			} else {
				set err [catch {file mkdir $destdir} errmsg]
				if {$err!=0} {
					tk_messageBox -title "PgcEdit: Save IFOs as" -icon error \
							-message "Could not create folder \"$destdir\":\n$errmsg" \
							-parent . -type ok
					return $err
				}
			}
		}
		#set ::config(initialdir) $destdir
		set errcode [::writedvd::writedvd $destdir]
		if {$errcode == 0} {
			tk_messageBox -title "PgcEdit: Save IFOs as" -icon warning \
					-message "IFOs saved OK!\nWARNING: VTS sectors calculated without VOBs!" \
					-parent . -type ok
		} elseif {$errcode == -1999} {
			tk_messageBox -title "PgcEdit: Save IFOs as" -icon warning \
					-message "IFOs saved OK!\nVTS sectors ERROR!" -parent . -type ok
		}
	} else {
		set errcode [::writedvd::writedvd $::dvddir]
		if {$errcode == 0} {
			tk_messageBox -title "PgcEdit: Save DVD" -icon info \
					-message "DVD saved OK!\nVTS sectors OK!" -parent . -type ok
			dvd_modified false false
		} elseif {$errcode == -1999} {
			tk_messageBox -title "PgcEdit: Save DVD" -icon warning \
					-message "DVD saved OK!\nVTS sectors ERROR!" -parent . -type ok
		}
	}
	return $errcode
}

#############################################################################
## Procedure:  Read_New_DVD
## read a new DVD in memory, and init the interface.
## Get input filename interactively if args is empty.
proc Read_New_DVD {args} {

	if $::dvd_modified {
		set r [tk_messageBox -message "The DVD has been modified.\n\nAre you sure?" \
				-icon warning -title "PgcEdit" -type yesno -default no]
		if {$r == "no"} {return 0}
	}

	if {$args == ""} {
		set initialdir $::config(initialdir)
		while true {
			set dvddir [tk_chooseDirectory -initialdir $initialdir \
				-parent . -title "PgcEdit: Open DVD" -mustexist true]
			if {$dvddir == ""} return
			if [file isdirectory [file join $dvddir "VIDEO_TS"]] {
				set dvddir [file join $dvddir "VIDEO_TS"]
			}
			set initialdir $dvddir
			if [file exists [file join $dvddir "VIDEO_TS.IFO"]] break
			bell
		}
	} else {
		set dvddir [file normalize [lindex $args 0]]
	}

	set errcode [readdvd $dvddir ::raw_sectors ::pgcs ::menus]
	if {$errcode == 0} {
		set ::dvddir $dvddir
		set ::config(initialdir) $dvddir
		if {$::config(auto_create_backup)} {
			Create_Backup true
		}
		# init GUI
		$::widget(main_cmd_lab) configure -text ""
		if {! $::scratchbook_active} {
			set ::scratchbook_active true
			scratchbook
		}
		set_window_title
		init_tree_menu

		display_pgc pgc 0 0

		# close any old information window
		destroy .streamsinfo
		destroy .dump

		tk_messageBox -title "PgcEdit: Open DVD" -icon info -parent . -type ok \
			-message "DVD Opened OK.\n\n$::pgcs(numvts) VTS loaded."
	} else {
		set ::dvddir x;	unset ::dvddir

		set backupdir [file join $dvddir PgcEdit_backup]
		if [file exists $backupdir] {
			set r [tk_messageBox -title "PgcEdit: Open DVD" -icon question \
				-message "Error opening DVD.\nRestaure backup?" -parent . \
				-type yesno -default yes]
			if {$r == "yes"} {
				file copy -force -- "$backupdir/VIDEO_TS.IFO" $dvddir
				file copy -force -- "$backupdir/VIDEO_TS.IFO" "$dvddir/VIDEO_TS.BUP"
				for {set vts 1} {$vts <= 1000} {incr vts} {
					if [file exists "$backupdir/VTS_[format %02i $vts]_0.IFO"] {
						file copy -force -- "$backupdir/VTS_[format %02i $vts]_0.IFO" $dvddir
						file copy -force -- "$backupdir/VTS_[format %02i $vts]_0.IFO" \
								"$dvddir/VTS_[format %02i $vts]_0.BUP"
					} else {
						break
					}
				}
				Read_New_DVD $dvddir
			}
		}
	}

	return $errcode
}

proc set_window_title {} {
	if {! [info exists ::dvddir]} return

	set txt ""

	set gn 	[::utils::get_General_Name]
	if {$gn != ""} {
		set txt "$txt - \"$gn\""
	} else {
		set txt "$txt - untitled"
	}

	set id [::utils::get_provider_ID]
	if {$id != ""} {
		set txt "${txt}; provider: $id"
	}

	if {$::dvd_modified} {
		set mod "*"
	} else {
		set mod " "
	}

	wm title . "PgcEdit - $mod \"$::dvddir/\"$txt"
}

proc dvd_modified {state onlyifnotscratchbook} {
	if {$onlyifnotscratchbook && $::scratchbook_active} return

	if {$state} {
		if {! $::dvd_modified} {
			set ::dvd_modified true
			set_window_title
		}
	} else {
		if {$::dvd_modified} {
			set ::dvd_modified false
			set_window_title
		}
	}
}

#############################################################################
## Procedure:  ReRead_DVD
## re-read the current DVD in memory, and loose changes.
proc ReRead_DVD {} {
	if [info exists ::dvddir] {
		set errcode [Read_New_DVD $::dvddir]
	} else {
		set errcode -9999
		bell
	}
	return $errcode
}

#############################################################################
## Procedure:  dump_DVD
## Dump all the DVD commands in a text string, and show the text.
proc dump_DVD {} {
	if [info exists ::dvddir] {
		set txt [::disasm::dump_all_commands "DVD \"$::dvddir/\""]
		showtext ".dump" "PgcEdit: DVD Dump" $txt
	} else {
		bell
	}
}

#############################################################################
## Procedure:  init_tree_menu
## Initialize the DVD menubutton with the DVD hierarchy stored in the ::pgcs
## and ::menus global arrays.
## arg: the tree menubutton
proc init_tree_menu {} {

	set old [.mf.f1.mb cget -menu]
	if {$old != ""} {
		destroy $old
	}
	set m .mf.f1.mb.menu
	menu $m
	.mf.f1.mb configure -menu $m

	set ::main::labelslist {}
	set ::main::cmdslist {}

	# VIDEO_TS.IFO
	$m add cascade -label "VIDEO_TS.IFO" -menu $m.vts0
	menu $m.vts0
	$m.vts0 add command -label "First-Play PGC" \
			-command "::main::display_pgc pgc 0 0"

	lappend ::main::labelslist "VMG / First-Play PGC"
	lappend ::main::cmdslist "::main::display_pgc pgc 0 0"

	for {set lu 1} {$lu <= $::menus(0,numlus)} {incr lu} {
		set lulabel "Menu-LU $lu ($::menus(0,$lu,lang))"
		if {$lu > 1 && [expr $lu % 20] == 1} {set break 1} else {set break 0}
		$m.vts0 add cascade -label $lulabel -menu $m.vts0.lu$lu -columnbreak $break
		menu $m.vts0.lu$lu
		for {set pgc 1} {$pgc <= $::menus(0,$lu,numpgcs)} {incr pgc} {
			set label "$pgc $::menus(0,$lu,$pgc,masktext)"
			if {$pgc > 1 && [expr $pgc % 20] == 1} {set break 1} else {set break 0}
			$m.vts0.lu$lu add command -label $label -columnbreak $break \
					-command "::main::display_pgc menu 0 $lu $pgc"

			lappend ::main::labelslist "VMG / $lulabel / $label"
			lappend ::main::cmdslist "::main::display_pgc menu 0 $lu $pgc"
		}
	}

	# VTS_xx_0.IFO
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		lappend ::main::labelslist ""
		lappend ::main::cmdslist "bell"

		set vtslabel "VTS_[format %02i $vts]_0.IFO"
		if {[expr $vts % 20] == 0} {set break 1} else {set break 0}
		$m add cascade -label $vtslabel -menu $m.vts$vts -columnbreak $break
		menu $m.vts$vts
		set numentries 0
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			incr numentries
			set lulabel "Menu-LU $lu ($::menus($vts,$lu,lang))"
			if {$numentries >= 20} {set break 1; set numentries 0} else {set break 0}
			$m.vts$vts add cascade -label $lulabel -columnbreak $break \
					-menu $m.vts$vts.lu$lu
			menu $m.vts$vts.lu$lu
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set pgclabel "$pgc $::menus($vts,$lu,$pgc,masktext)"
				if {$pgc > 1 && [expr $pgc % 20] == 1} {set break 1} else {set break 0}
				$m.vts$vts.lu$lu add command -label $pgclabel -columnbreak $break \
					-command "::main::display_pgc menu $vts $lu $pgc"

				lappend ::main::labelslist "VTS $vts / $lulabel / $pgclabel"
				lappend ::main::cmdslist "::main::display_pgc menu $vts $lu $pgc"
			}
		}
		if {$::pgcs($vts,numpgcs) > 0} {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				incr numentries
				if {$numentries >= 20} {set break 1; set numentries 0} else {set break 0}
				set label "TTN $pgc"
				if [info exists ::pgcs($vts,$pgc,Title)] {
					set label "$label  (Title $::pgcs($vts,$pgc,Title))"
				}
				$m.vts$vts add command -label $label -columnbreak $break \
					-command "::main::display_pgc pgc $vts $pgc"

				lappend ::main::labelslist "VTS $vts / $label"
				lappend ::main::cmdslist "::main::display_pgc pgc $vts $pgc"
			}
		}
	}
}

#############################################################################
## Procedure:  selector
## mode arg: "show", "hide" or "toggle"
proc selector {mode labellist cmdslist} {
	set w .mf.f2.selector

	if {! [info exists $labellist]} {
		set $labellist {}
		set $cmdslist {}
	}

	switch $mode {
		"show" {
			set show 1
		}
		"hide" {
			set show 0
		}
		"toggle" {
			if [winfo exists $w.a] {
				set show 0
			} else {
				set show 1
			}
		}
	}
	if $show {
		set ::main::selector_opened 1

		frame $w.a
		listbox   $w.a.lb -exportselection 0 -selectmode browse \
				-listvariable $labellist -width 0 \
				-highlightcolor gray -activestyle dotbox \
				-yscrollcommand "$w.a.ys set" \
				;#-xscrollcommand "$w.a.xs set"
		scrollbar $w.a.ys -command "$w.a.lb yview"
		#scrollbar $w.a.xs -command "$w.a.lb xview" -orient horizontal
		frame     $w.a.pad -width 2
		grid  $w.a.lb -column 0 -row 0 -sticky nesw
		grid  $w.a.ys -column 1 -row 0 -sticky ns
		#grid  $w.a.xs -column 0 -row 1 -sticky ew
		grid  $w.a.pad -column 2 -row 0 -sticky ew
		grid columnconf $w.a 0 -weight 1
		grid rowconf    $w.a 0 -weight 1

		bind $w.a.lb <1> "focus $w.a.lb; ::main::selector_launch_cmd_click $w $cmdslist"
		bind $w.a.lb <2> "focus $w.a.lb; ::main::selector_launch_cmd_click $w $cmdslist; ::main::action_Edit_PGC"
		bind $w.a.lb <3> "focus $w.a.lb; ::main::selector_launch_cmd_click $w $cmdslist; ::main::action_Edit_PGC"
		bind $w.a.lb <Return> "::main::selector_launch_cmd_key $w $cmdslist"
		bind $w.a.lb <space>  "::main::selector_launch_cmd_key $w $cmdslist"
		bind $w.a.lb <Home> "$w.a.lb see 0; $w.a.lb activate 0; $w.a.lb selection clear 0 end; $w.a.lb selection set 0"
		bind $w.a.lb <End>  "$w.a.lb see end; $w.a.lb activate end; $w.a.lb selection clear 0 end; $w.a.lb selection set end"
		bind $w.a.lb <MouseWheel> "if {%D>0} {$w.a.lb yview scroll 1 units} else {$w.a.lb yview scroll -1 units}"

		pack  $w.a -side top -fill both -expand 1
	} else {
		set ::main::selector_opened 0

		#bind . <Destroy> {}
		destroy $w.a
		#bind . <Destroy> {cleanexit 0}
		$w configure -width 1	;# don't work with 0
	}
}
proc selector_launch_cmd_click {w cmdslist} {
	set idx [$w.a.lb index @[expr [winfo pointerx .]-[winfo rootx $w.a.lb]],[expr [winfo pointery .]-[winfo rooty $w.a.lb]]]
	$w.a.lb activate $idx
	$w.a.lb selection clear 0 end
	$w.a.lb selection set $idx
	eval [lindex [subst $$cmdslist] [$w.a.lb index $idx]]
}
proc selector_launch_cmd_key {w cmdslist} {
	eval [lindex [subst $$cmdslist] [$w.a.lb index active]]
}

#############################################################################
## Procedure:  refresh_listbox
## refresh the listbox with the current listbox contents: re-disassemble all
## commands, regenerate line numbers...
proc refresh_listbox {} {
	set cmdnum 0
	set linenum 1
	foreach cmd $::editor_cmds_list {
		switch [string range $cmd 0 0] {
			{*} {
				set linenum 1
				set newcmd [formatCommand "* [lindex $cmd 1]" 0]
				$::widget(main_listbox) itemconfigure $cmdnum -background $::config(systembuttonface)
			}
			{[} {
				set newcmd [formatCommand [string range $cmd 1 23] $linenum]
				incr linenum
				if {[string first "ILLEGAL" $newcmd] >= 0} {
					$::widget(main_listbox) itemconfigure $cmdnum -background pink
					$::widget(main_listbox) itemconfigure $cmdnum -selectbackground orange
				} else {
					$::widget(main_listbox) itemconfigure $cmdnum -background white
					$::widget(main_listbox) itemconfigure $cmdnum -selectbackground $::config(systemhighlight)
				}
			}
			default {
				set newcmd [formatCommand $cmd $linenum]
				incr linenum
				if {[string first "ILLEGAL" $newcmd] >= 0} {
					$::widget(main_listbox) itemconfigure $cmdnum -background pink
					$::widget(main_listbox) itemconfigure $cmdnum -selectbackground orange
				} else {
					$::widget(main_listbox) itemconfigure $cmdnum -background white
					$::widget(main_listbox) itemconfigure $cmdnum -selectbackground $::config(systemhighlight)
				}
			}
		}
		lset ::editor_cmds_list $cmdnum $newcmd
		incr cmdnum
	}
}

#############################################################################
## Procedure:  display_pgc
## Display a PGC in the listbox, and update the PGC label beside the tree menubutton
## usage:
##  display_pgc "pgc" vtsnum pgcnum
##  display_pgc "menu" vtsnum lunum pgcnum
proc display_pgc {type vts args} {

	if {$vts == 0} {
		set vtslabel "\"VIDEO_TS.IFO\":   VMG"
	} else {
		set vtslabel "\"VTS_[format %02i $vts]_0.IFO\":   VTS $vts"
	}

	if {$type == "pgc"} {
		set pgc $args
		if {$vts == 0} {
			set pgclabel "First Play PGC"
		} else {
			if [info exists ::pgcs($vts,$pgc,Title)] {
				set pgclabel "TTN $pgc  (Title $::pgcs($vts,$pgc,Title))"
			} else {
				set pgclabel "TTN $pgc"
			}
		}
		cmds_to_listbox pgc $vts,$pgc "$vtslabel / $pgclabel"

	} else {
		set lu  [lindex $args 0]
		set pgc [lindex $args 1]
		set lulabel "Menu-LU $lu ($::menus($vts,$lu,lang))"
		set pgclabel "$pgc $::menus($vts,$lu,$pgc,masktext)"
		cmds_to_listbox menu $vts,$lu,$pgc "$vtslabel / $lulabel / $pgclabel"
	}
	#if {[winfo exists .pgceditor] && [winfo ismapped .pgceditor]} action_Edit_PGC
}

#############################################################################
## Procedure:  cmds_to_listbox
## init the global variable ::editor_cmds_list with the current commands
proc cmds_to_listbox {type index label} {

	$::widget(main_cmd_lab) configure -text $label

	if $::scratchbook_active {
		set ::scratchbook_old_selection [get_selection_indexes]
		set ::scratchbook_active false
	}

	# remember from where the commands are coming from
	set ::editor_cmds_from [list $type $index]

	# init editor list
	switch $type {
		pgc {
			set ::editor_cmds_list   $::pgcs($index)
			set ::editor_undo_buffer $::pgcs($index)
		}
		menu {
			set ::editor_cmds_list   $::menus($index)
			set ::editor_undo_buffer $::menus($index)
		}
	}
	refresh_listbox
	set_listbox_selection 0
}

#############################################################################
## Procedure:  listbox_to_cmds
## copies the listbox contents from $::editor_cmds_list to the original part
## of the DVD $::pgcs($index) or $::menus($index)
proc listbox_to_cmds {} {
	set rawcmds {}
	foreach line $::editor_cmds_list {
		if {[string range $line 0 0] == "*"} {
			lappend rawcmds "* [lindex $line 1]"
		} else {
			lappend rawcmds "[string range $line 1 23]"
		}
	}
	if $::scratchbook_active {
		set ::scratchbook_buffer $rawcmds
	} else {
		set type  [lindex $::editor_cmds_from 0]
		set index [lindex $::editor_cmds_from 1]
		set [subst ::${type}s($index)] $rawcmds
	}
}

#############################################################################
## Procedure:  scratchbook
## Configure the listbox for the scratchbook or the normal DVD commands upon
## state of the bool $::scratchbook_active
proc scratchbook {} {
	set oldselection $::scratchbook_old_selection
	if {[lindex $oldselection 0] == "-"} {set oldselection {0 0}}
	if $::scratchbook_active {
		set ::scratchbook_old_selection [get_selection_indexes]
		set ::editor_cmds_list $::scratchbook_buffer
		set_listbox_selection [lindex $oldselection 0] [lindex $oldselection 1]

	} else {
		if [info exists ::editor_cmds_from] {
			set ::scratchbook_old_selection [get_selection_indexes]
			set type  [lindex $::editor_cmds_from 0]
			set index [lindex $::editor_cmds_from 1]
			set ::editor_cmds_list [subst \$::${type}s($index)]
			set_listbox_selection [lindex $oldselection 0] [lindex $oldselection 1]
		} else {
			set ::scratchbook_active true
			bell
		}
	}
	refresh_listbox
}
proc save_scratchbook {} {
	if {[llength $::scratchbook_buffer] < 2} {
		bell
		return
	}
	set filename [tk_getSaveFile -filetypes {{"scratchbook files" {*.sbk}}} \
		-initialdir "[file dirname $::config(scratchbook_initialfile)]" \
		-initialfile "[file tail $::config(scratchbook_initialfile)]" \
		-title "PgcEdit: Save scratchbook"]
	if {$filename == ""} {
		return
	}
	if {[file extension $filename] != ".sbk"} {
		set filename "${filename}.sbk"
	}
	set ::config(scratchbook_initialfile) $filename
	saveFile $filename binary $::scratchbook_buffer
}
proc open_scratchbook {} {
	set filename [tk_getOpenFile -filetypes {{"scratchbook files" {*.sbk}}} \
		-initialdir "[file dirname $::config(scratchbook_initialfile)]" \
		-initialfile "[file tail $::config(scratchbook_initialfile)]" \
		-title "PgcEdit: Open scratchbook"]
	if {$filename == ""} {
		return
	}
	if {[file extension $filename] != ".sbk"} {
		set filename "${filename}.sbk"
	}
	set ::config(scratchbook_initialfile) $filename
	set buf [readFile $filename binary]
	if [catch {set line0 [lindex $buf 0]}] {
		bell
		return
	}
	if {[lindex $line0 1] != "scratchbook"} {
		bell
		return
	}
	set ::scratchbook_buffer $buf
	set ::scratchbook_active true
	scratchbook
}

#############################################################################
## Procedure:  set_listbox_selection
## set the selection of the listbox to the provided range
## Args: index1 ?index2? (if index2 is not provided, select only one line)
proc set_listbox_selection {idx1 args} {
	if {$args == ""} {set idx2 $idx1} else {set idx2 $args}
	$::widget(main_listbox) activate $idx1
	$::widget(main_listbox) selection clear 0 end
	$::widget(main_listbox) selection set $idx1 $idx2
	$::widget(main_listbox) see $idx2
	$::widget(main_listbox) see $idx1
}

#############################################################################
## Procedure:  get_selection_indexes
## retreave the first and last indexes of the selection in the listbox.
## return a list with the two indexes, or {- -} if nothing is selected.
## Note: the two indexes are identical if only one command is selected.
## Do NOT check for selection of pre, post or cell line.
proc get_selection_indexes {} {
	set selection [$::widget(main_listbox) curselection]
	if {$selection == ""} {
		return [list - -]
	}
	return [list [lindex $selection 0] \
				 [lindex $selection [expr [llength $selection]-1]] ]
}

#############################################################################
## Procedure:  get_selected_cmds
## get the selected commands from the listbox
## return a list with the commands in hex form, or bell if none is selected,
## or if a pre, post or cell label is in the selection.
proc get_selected_cmds {} {
	set idxs [get_selection_indexes]
	set idx1 [lindex $idxs 0]
	set idx2 [lindex $idxs 1]
	if {$idx1 == "-"} {
		bell
		return ""
	}
	set cmds {}
	for {set cidx $idx1} {$cidx <= $idx2} {incr cidx} {
		set cmdtext [$::widget(main_listbox) get $cidx]
		if {[string range $cmdtext 0 0] != {[} } {
			bell
			return ""
		}
		lappend cmds [string range $cmdtext 1 23]
	}
	return $cmds
}

#############################################################################
## Procedure:  delete_selected_cmds
## delete the selected commands.
## Return the original starting index of the selection, or "-" for error.
proc delete_selected_cmds {} {
	set idxs [get_selection_indexes]
	set idx1 [lindex $idxs 0]
	set idx2 [lindex $idxs 1]
	if {$idx1 == "-"} {
		bell
		return "-"
	}
	set ::editor_cmds_list [lreplace $::editor_cmds_list $idx1 $idx2]
	return $idx1
}

#############################################################################
## Procedure:  insert_cmds
## insert commands in the listbox, just before the provided index.
proc insert_cmds {index commands} {
	if {[expr [llength $::editor_cmds_list]+[llength $commands]] > 131} {
		bell
		return -1
	}
	if {$index == 0} {
		set ::editor_cmds_list [concat $commands $::editor_cmds_list]
	} elseif {$index >= [llength $::editor_cmds_list]} {
		set ::editor_cmds_list [concat $::editor_cmds_list $commands]
	} else {
		set ::editor_cmds_list [concat \
				[lrange $::editor_cmds_list 0 [expr $index -1]] \
				$commands \
				[lrange $::editor_cmds_list $index end] ]
	}
	return [llength $::editor_cmds_list]
}

#############################################################################
## highlight goto line
proc highlight_goto {linenum} {
	set line [lindex $::editor_cmds_list $linenum]
	set cmd [string range $line 0 23]
	set cmdchk "[string range $cmd 1 2][string range $cmd 5 5]"
	if {$cmdchk=="001" || $cmdchk=="003"} {
		set err 0
		set ln [expr 0x[string range $cmd 22 23]]
		if {$ln>0 && $ln<129} {
			set highln [expr $linenum + [expr $ln - [lindex $line 8]]]
			if {![catch {set origcolor [$::widget(main_listbox) itemcget $highln -background]}]} {
				set destline [lindex $::editor_cmds_list $highln]
				if {[string range $destline 0 0]=="\["} {
					if {[lindex $destline 8]==$ln} {
						$::widget(main_listbox) itemconfigure $highln -background lightblue
						after 333 "$::widget(main_listbox) itemconfigure $highln -background $origcolor"
					} else {
						set err 1
					}
				} else {
					set err 1
				}
			} else {
				set err 1
			}
		} else {
			set err 1
		}
		if $err {
			$::widget(main_listbox) itemconfigure $linenum -selectbackground red
			after 333 "$::widget(main_listbox) itemconfigure $linenum -selectbackground $::config(systemhighlight)"
		}
	}
}
#############################################################################
## highlight goto line
## command bound to mouse button 1 on a line of the main listbox
proc highlight_goto_bound {} {
	highlight_goto [.mf.f2.main.l index @[expr [winfo pointerx .]-[winfo rootx .mf.f2.main.l]],[expr [winfo pointery .]-[winfo rooty .mf.f2.main.l]]]
}

#############################################################################
## Action Procedures
## procs called by the action buttons (below the listbox)
## There are no args to theses commands

proc action_Edit_Cmd {} {
	set idxs [get_selection_indexes]
	set idx [lindex $idxs 0]
	if {$idx!=[lindex $idxs 1]} {
		bell
		return
	}
	set cmd [get_selected_cmds]
	if {$cmd == ""} {
		bell
		return
	}
	set ::hexcommand [lindex $cmd 0]

	# we use this variable to know when the command editor returns
	set ::ced_return -1
	::ced::cmdeditor
	tkwait variable ::ced_return
	if {$::ced_return == 1} {
		lset ::editor_cmds_list $idx $::hexcommand
		refresh_listbox
		listbox_to_cmds
		highlight_goto $idx
		dvd_modified true true
	}
}

proc action_Edit_PGC {} {
	if {! [info exists ::dvddir] || $::scratchbook_active} {
		bell
		return
	}
	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]
	set tmp ::${type}s($index,header)
	set pgcheader [subst $$tmp]
	set tmp ::${type}s($index,numprograms)
	set numprograms [subst $$tmp]
	set tmp ::${type}s($index,programmap)
	set programmap [subst $$tmp]
	set tmp ::${type}s($index,numcells)
	set numcells [subst $$tmp]
	set tmp ::${type}s($index,cellpbtable)
	set cellpbtable [subst $$tmp]
	set tmp ::${type}s($index,cellpostable)
	set cellpostable [subst $$tmp]

	# If it is a title, read the PTT infos in the PTT_SRPT table
	# so that we can display the chapter number in the Cell table.
	set PTTlist ""
	if {$type == "pgc"} {
		set vtsnum [lindex [split $index ","] 0]
		set ttnnum [lindex [split $index ","] 1]
		if {$vtsnum > 0} {
			set data $::raw_sectors($vtsnum,VTS_PTT_SRPT)
			if {$data != ""} {
				set numttus [::readdvd::get_int data 0 2]
				if {$numttus >= $ttnnum} {
					set idx [::readdvd::get_int data [expr 4 + ($ttnnum * 4)] 4]
					set pgcn [::readdvd::get_int data $idx 2]
					while {$pgcn == $ttnnum} {
						incr idx 2
						lappend PTTlist [::readdvd::get_int data $idx 2]
						incr idx 2
						if {$idx >= [string length $data]} break
						set pgcn [::readdvd::get_int data $idx 2]
					}
				}
			}
		}
	}

	set r [::pgced::edit_pgc $pgcheader $numprograms $programmap \
			$numcells $cellpbtable $cellpostable $PTTlist]
	if {$r != ""} {
		set ::${type}s($index,header)      [lindex $r 0]
		set ::${type}s($index,cellpbtable) [lindex $r 1]
		dvd_modified true true
	}
}

proc action_Undo_pgc {} {
	if [info exists ::editor_undo_buffer] {
		set ::editor_cmds_list $::editor_undo_buffer
		refresh_listbox
		set_listbox_selection 0
		listbox_to_cmds
	} else {
		bell
	}
}

proc action_Delete {} {
	if {[get_selected_cmds] != ""} {
		set newidx [delete_selected_cmds]
		if {$newidx != ""} {
			refresh_listbox
			set_listbox_selection $newidx
			if {[lindex [get_selection_indexes] 0] == "-"} {
				set_listbox_selection end
			}
			listbox_to_cmds
			dvd_modified true true
		}
	}
}

proc action_Duplicate {} {
	set newcmds [get_selected_cmds]
	if {$newcmds != ""} {
		set idxs [get_selection_indexes]
		set idx1 [lindex $idxs 0]
		set idx2 [lindex $idxs 1]
		if {[insert_cmds $idx1 $newcmds] != -1} {
			refresh_listbox
			set_listbox_selection [expr $idx2 + 1] [expr $idx2 + 1 + $idx2 - $idx1]
			listbox_to_cmds
			dvd_modified true true
		}
	}
}

proc action_Add_before {} {
	set idx [lindex [get_selection_indexes] 0]
	if {$idx != "-" && $idx != 0} {
		if {[insert_cmds $idx {{00 00 00 00 00 00 00 00}}] != -1} {
			refresh_listbox
			set_listbox_selection $idx $idx
			listbox_to_cmds
			dvd_modified true true
		}
	} else {
		bell
	}
}

proc action_Add_after {} {
	set idx [lindex [get_selection_indexes] 1]
	if {$idx != "-"} {
		incr idx 1
		if {[insert_cmds $idx {{00 00 00 00 00 00 00 00}}] != -1} {
			refresh_listbox
			set_listbox_selection $idx $idx
			listbox_to_cmds
			dvd_modified true true
		}
	} else {
		bell
	}
}

proc action_Move_up {} {
	if {[get_selected_cmds] != ""} {
		set idxs [get_selection_indexes]
		set idx1 [lindex $idxs 0]
		set idx2 [lindex $idxs 1]
		if {$idx1 < 2} {
			bell
			return
		}
		set part1  [lrange $::editor_cmds_list 0 [expr $idx1-2]]
		set cutcmd [lrange $::editor_cmds_list [expr $idx1-1] [expr $idx1-1]]
		set part2  [lrange $::editor_cmds_list $idx1 $idx2]
		set part3  [lrange $::editor_cmds_list [expr $idx2+1] end]
		set ::editor_cmds_list [concat $part1 $part2 $cutcmd $part3]
		refresh_listbox
		set_listbox_selection [expr $idx1-1] [expr $idx2-1]
		listbox_to_cmds
		dvd_modified true true
	}
}

proc action_Move_down {} {
	if {[get_selected_cmds] != ""} {
		set idxs [get_selection_indexes]
		set idx1 [lindex $idxs 0]
		set idx2 [lindex $idxs 1]
		if {$idx2 >= [expr [llength $::editor_cmds_list]-1]} {
			bell
			return
		}
		set part1  [lrange $::editor_cmds_list 0 [expr $idx1-1]]
		set part2  [lrange $::editor_cmds_list $idx1 $idx2]
		set cutcmd [lrange $::editor_cmds_list [expr $idx2+1] [expr $idx2+1]]
		set part3  [lrange $::editor_cmds_list [expr $idx2+2] end]
		set ::editor_cmds_list [concat $part1 $cutcmd $part2 $part3]
		refresh_listbox
		set_listbox_selection [expr $idx1+1] [expr $idx2+1]
		listbox_to_cmds
		dvd_modified true true
	}
}

proc action_Copy {} {
	set cmds [get_selected_cmds]
	if {$cmds != ""} {
		clipboard clear
		clipboard append -- $cmds
	}
}

proc action_Append {} {
	set cmds [get_selected_cmds]
	if {$cmds != ""} {
		set oldcmds [action_Paste_get_clipboard]
		if {$oldcmds != ""} {
			clipboard clear
			clipboard append -- [concat $oldcmds $cmds]
		}
	}
}

proc action_Cut {} {
	set cmds [get_selected_cmds]
	if {$cmds != ""} {
		clipboard clear
		clipboard append -- $cmds
		action_Delete
		dvd_modified true true
	}
}

proc action_Paste_get_clipboard {} {
	if {[catch {set cmds [clipboard get]}]} {
		bell
		return ""
	}
	if {[string range $cmds 0 0] != "\{" && [string length $cmds] == 23} {
		set cmds "\{$cmds\}"
	}
	if {[string range $cmds 0 0] != "\{"} {
		bell
		return ""
	}
	return $cmds
}

proc action_Paste_before {} {
	set cmds [action_Paste_get_clipboard]
	if {$cmds == ""} {return}
	if {[expr [llength $::editor_cmds_list]+[llength $cmds]] > 131} {
		bell
		return
	}
	set idx [lindex [get_selection_indexes] 0]
	if {$idx != "-" && $idx != 0} {
		set part1 [lrange $::editor_cmds_list 0 [expr $idx-1]]
		set part2 [lrange $::editor_cmds_list $idx end]
		set ::editor_cmds_list [concat $part1 $cmds $part2]
		refresh_listbox
		set_listbox_selection $idx [expr $idx+[llength $cmds]-1]
		listbox_to_cmds
		dvd_modified true true
	} else {
		bell
	}
}

proc action_Paste_after {} {
	set cmds [action_Paste_get_clipboard]
	if {$cmds == ""} {return}
	if {[expr [llength $::editor_cmds_list]+[llength $cmds]] > 131} {
		bell
		return
	}
	set idx [lindex [get_selection_indexes] 1]
	if {$idx != "-"} {
		set part1 [lrange $::editor_cmds_list 0 $idx]
		set part2 [lrange $::editor_cmds_list [expr $idx+1] end]
		set ::editor_cmds_list [concat $part1 $cmds $part2]
		refresh_listbox
		set_listbox_selection [expr $idx+1] [expr $idx+[llength $cmds]]
		listbox_to_cmds
		dvd_modified true true
	} else {
		bell
	}
}

proc action_Paste_replace {} {
	set cmds [action_Paste_get_clipboard]
	if {$cmds == ""} {return}

	if {[get_selected_cmds] != ""} {
		set idxs [get_selection_indexes]
		set numselected [expr [lindex $idxs 1]-[lindex $idxs 0]+1]
		if {[expr [llength $::editor_cmds_list]-$numselected+[llength $cmds]] > 131} {
			bell
			return
		}
		set idx [delete_selected_cmds]
		if {$idx != ""} {
			set part1 [lrange $::editor_cmds_list 0 [expr $idx-1]]
			set part2 [lrange $::editor_cmds_list $idx end]
			set ::editor_cmds_list [concat $part1 $cmds $part2]
			refresh_listbox
			set_listbox_selection $idx [expr $idx+[llength $cmds]-1]
			listbox_to_cmds
			dvd_modified true true
		}
	}
}

proc action_Mod_GOTOs {increment} {
	set idxs [get_selection_indexes]
	set idx1 [lindex $idxs 0]
	set idx2 [lindex $idxs 1]
	if {$idx1 == "-"} {
		bell
		return
	}
	set modified false
	for {set idx $idx1} {$idx <= $idx2} {incr idx} {
		set cmd [string range [lindex $::editor_cmds_list $idx] 0 23]
		set cmdchk "[string range $cmd 1 2][string range $cmd 5 5]"
		if {$cmdchk == "001" || $cmdchk == "003"} {
			set ln [expr 0x[string range $cmd 22 23]]
			if {$cmdchk == "003" && $ln == 0} continue
			incr ln $increment
			if {$cmdchk == "003" && $ln == 0} continue
			if {$ln <= 0 || $ln > 128} {
				bell
			} else {
				lset ::editor_cmds_list $idx \
						"\[[string range $cmd 1 20] [format %02X $ln]\]"
				set modified true
			}
		}
	}
	if {$modified} {
		refresh_listbox
		listbox_to_cmds
		dvd_modified true true
	}
}

proc action_Incr_GOTOs {} {
	action_Mod_GOTOs 1
}

proc action_Decr_GOTOs {} {
	action_Mod_GOTOs -1
}

## action buttons and/or popup menu
## mode arg: "show", "hide", "toggle" or "init"
proc action_buttons {mode} {
	set cmdlist {
		"Edit Cmd"
		"Edit PGC"
		"Add before"
		"Add after"
		"Delete"
		"Duplicate"
		"Copy"
		"Append"
		"Cut"
		"Paste replace"
		"Paste before"
		"Paste after"
		"Decr GOTOs"
		"Incr GOTOs"
		"Move up"
		"Move down"
	}
	if {! [winfo exists .actionmenu]} {
		menu .actionmenu
		if {$::config(tearoff_menus)} {
			set menunum 1
		} else {
			set menunum 0
		}
		set ::main::last_menu_selected $menunum
		foreach cmd $cmdlist {
			if {$cmd != "-"} {
				set cmdname "::main::action_[string map {" " "_"} $cmd]"
				.actionmenu add command \
						-label $cmd -command "set ::main::last_menu_selected $menunum; $cmdname"
				incr menunum
			}
		}
	}

	switch $mode {
		"show" {
			set show 1
		}
		"hide" {
			set show 0
		}
		"toggle" {
			if [winfo exists .mf.cf.f] {
				set show 0
			} else {
				set show 1
			}
		}
		"init" {
			if {$::config(buttons)} {
				set show 1
			} else {
				set show 0
			}
		}
	}
	if $show {
		set ::main::buttons_opened 1

		if {! [winfo exists .mf.cf.f]} {
			frame .mf.cf.f
			set col 0
			set row 0
			foreach cmd $cmdlist {
				if {$cmd != "-"} {
					set cmdname "::main::action_[string map {" " "_"} $cmd]"
					button .mf.cf.f.b${row}_$col -text $cmd -command $cmdname -width 10 -pady 0
					grid .mf.cf.f.b${row}_$col -column $col -row $row -sticky ew -padx 1
				}
				if {$row == 0} {
					incr row
					grid columnconf .mf.cf.f $col -weight 1
				} else {
					set row 0
					incr col
				}
				grid rowconf .mf.cf.f 0 -pad 2
				grid rowconf .mf.cf.f 1 -pad 2
			}
			pack .mf.cf.f -fill x -pady 1
		}
	} else {
		set ::main::buttons_opened 0

		if [winfo exists .mf.cf.f] {
			#bind . <Destroy> {}
			destroy .mf.cf.f
			#bind . <Destroy> {cleanexit 0}
			.mf.cf configure -height 1	;# don't work with 0
		}
	}
}

#############################################################################
## Main program entry
#############################################################################

wm withdraw .
update idletasks

#bind . <Destroy> {cleanexit 0}
wm protocol . WM_DELETE_WINDOW {cleanexit 0}

namespace import ::readdvd::*
namespace import ::disasm::*

set ::scratchbook_old_selection {0 0}
set ::scratchbook_active true
set ::scratchbook_buffer [list {* scratchbook}]
set ::dvd_modified false

#################
#
# Build GUI.
#

# the DVD commands as they appear in the listbox window
set ::editor_cmds_list {}

frame .mf
pack  .mf -fill both -expand 1 -padx 3 -pady 2
frame .mf.f1
pack  .mf.f1 -fill x -pady 1 -padx 1
menubutton .mf.f1.mb -text "PGC" -pady 0 -width 12
pack  .mf.f1.mb -side left -fill y
bind  .mf.f1.mb <3> "::main::selector toggle ::main::labelslist ::main::cmdslist"
bind  .mf.f1.mb <Control-Double-2> {console show}
label .mf.f1.lab -text "" -anchor center -relief sunken -bd 1
pack  .mf.f1.lab -side left -padx 6 -pady 1 -fill both -expand 1
set ::widget(main_cmd_lab) .mf.f1.lab
checkbutton .mf.f1.sb -text "Scratchbook" -command ::main::scratchbook \
	-variable ::scratchbook_active -onvalue true -offvalue false -width 12 \
	-indicatoron false -relief sunken -offrelief raised -selectcolor pink -pady 0
pack .mf.f1.sb -side right -pady 1 -fill y

# main listbox
frame .mf.f2
pack  .mf.f2 -fill both -expand 1 -pady 1 -padx 1
frame .mf.f2.selector
pack  .mf.f2.selector -fill both -expand 0 -side left
frame .mf.f2.main
pack  .mf.f2.main -fill both -expand 1 -side left
listbox .mf.f2.main.l -exportselection 0 -selectmode extended \
		-xscrollcommand ".mf.f2.main.xscr set" -yscrollcommand ".mf.f2.main.yscr set" \
		-listvariable ::editor_cmds_list -font $::config(fixed_width_font) \
		-activestyle none -highlightcolor gray60 -height 3
scrollbar .mf.f2.main.yscr -command ".mf.f2.main.l yview"
scrollbar .mf.f2.main.xscr -command ".mf.f2.main.l xview" -orient horizontal
grid  .mf.f2.main.l    -column 0 -row 0 -sticky nesw
grid  .mf.f2.main.yscr -column 1 -row 0 -sticky ns
grid  .mf.f2.main.xscr -column 0 -row 1 -sticky ew
grid columnconf .mf.f2.main 0 -weight 1
grid rowconf    .mf.f2.main 0 -weight 1

# action buttons
frame .mf.cf
pack  .mf.cf -fill x
action_buttons init
bind .mf.f2.main.l <3> {tk_popup .actionmenu [winfo pointerx .] [winfo pointery .] $::main::last_menu_selected}
bind .mf.f2.main.l <Control-1>     {.mf.f2.main.l selection clear 0 end}
bind .mf.f2.main.l <1> {focus .mf.f2.main.l; ::main::highlight_goto_bound}
bind .mf.f2.main.l <Double-1>      ::main::action_Edit_Cmd
bind .mf.f2.main.l <2>             ::main::action_Edit_PGC
bind .mf.f2.main.l <space>         ::main::action_Edit_Cmd
bind .mf.f2.main.l <Return>        ::main::action_Edit_PGC
bind .mf.f2.main.l <Control-z>     ::main::action_Undo_pgc
bind .mf.f2.main.l <Delete>        ::main::action_Delete
bind .mf.f2.main.l <Control-d>     ::main::action_Duplicate
bind .mf.f2.main.l <Control-less>  ::main::action_Add_before
bind .mf.f2.main.l <Control-greater> ::main::action_Add_after
bind .mf.f2.main.l <Insert>        ::main::action_Add_after
bind .mf.f2.main.l <Control-bracketleft> ::main::action_Move_up
bind .mf.f2.main.l <Control-bracketright> ::main::action_Move_down
bind .mf.f2.main.l <Control-c>     ::main::action_Copy
bind .mf.f2.main.l <Control-x>     ::main::action_Cut
bind .mf.f2.main.l <Control-v>     ::main::action_Paste_replace
bind .mf.f2.main.l <Control-plus>  ::main::action_Incr_GOTOs
bind .mf.f2.main.l <Control-minus> ::main::action_Decr_GOTOs
bind .mf.f2.main.l <Control-a> {::main::set_listbox_selection 0 end}
bind .mf.f2.main.l <Escape> {.mf.f2.main.l see 0; .mf.f2.main.l activate 0; .mf.f2.main.l selection set 0}
bind .mf.f2.main.l <Home>   {.mf.f2.main.l see 0; .mf.f2.main.l activate 0; .mf.f2.main.l selection clear 0 end; .mf.f2.main.l selection set 0}
bind .mf.f2.main.l <End>    {.mf.f2.main.l see end; .mf.f2.main.l activate end; .mf.f2.main.l selection clear 0 end; .mf.f2.main.l selection set end}
bind .mf.f2.main.l <MouseWheel> {
	if {%D > 0} {
		.mf.f2.main.l yview scroll  1 units
	} else {
		.mf.f2.main.l yview scroll -1 units
	}
}

set ::widget(main_listbox) .mf.f2.main.l


# main menu
menu .mainmenu

.mainmenu add cascade -label "File" -menu .mainmenu.file
menu .mainmenu.file
.mainmenu.file add command -label "Open DVD..." -command ::main::Read_New_DVD -accelerator "Ctrl+O"
bind . <Control-o> ::main::Read_New_DVD
.mainmenu.file add command -label "Save DVD..." -command {::main::Write_DVD false} -accelerator "Ctrl+S"
bind . <Control-s> {::main::Write_DVD false}
.mainmenu.file add command -label "Reopen DVD" -command ::main::ReRead_DVD -accelerator "Ctrl+R"
bind . <Control-r> ::main::ReRead_DVD
.mainmenu.file add separator
.mainmenu.file add command -label "Save IFOs as..." -command {::main::Write_DVD true}
.mainmenu.file add separator
.mainmenu.file add command -label "Check DVD+/-R Size..." -command {::writedvd::DVD_size}
.mainmenu.file add separator
.mainmenu.file add command -label "Create backup"  -command {::main::Create_Backup false} -accelerator "Ctrl+B"
bind . <Control-b> {::main::Create_Backup false}
.mainmenu.file add command -label "Restore backup" -command ::main::Restore_Backup
.mainmenu.file add command -label "Delete backup"  -command ::main::Delete_Backup
.mainmenu.file add separator
.mainmenu.file add command -label "Open scratchbook..." -command ::main::open_scratchbook
.mainmenu.file add command -label "Save scratchbook..." -command ::main::save_scratchbook
.mainmenu.file add separator
.mainmenu.file add command -label "Dump DVD commands..." -command ::main::dump_DVD
if {$::config(debug) > 0} {
	.mainmenu.file add separator
	.mainmenu.file add command -label "Console" -command {console show}
}
.mainmenu.file add separator
.mainmenu.file add command -label "Quit" -command {cleanexit 0}

.mainmenu add cascade -label "Edit" -menu .actionmenu

.mainmenu add cascade -label "Utilities" -menu .mainmenu.utils
::utils::utils_menu .mainmenu.utils

.mainmenu add cascade -label "Tools" -menu .mainmenu.tools
menu .mainmenu.tools
::toolscfg::configure_tools_menu

.mainmenu add cascade -label "View" -menu .mainmenu.view
menu .mainmenu.view
.mainmenu.view add checkbutton -label "PGC Selector" \
		-variable ::main::selector_opened \
		-command {::main::selector toggle ::main::labelslist ::main::cmdslist}
.mainmenu.view add checkbutton -label "Edit Buttons" \
		-variable ::main::buttons_opened \
		-command {::main::action_buttons toggle}
.mainmenu.view add separator
.mainmenu.view add command -label "Tcl Calculator/Converter" \
		-command {::calc::calculator .calc}

.mainmenu add cascade -label "Options" -menu .mainmenu.options
menu .mainmenu.options

if {$::tcl_platform(platform) == "windows" && [info exists ::freewrap::progname]} {
	.mainmenu.options add cascade -label "Install" -menu .mainmenu.options.install
	.mainmenu.options add separator

	menu .mainmenu.options.install
	.mainmenu.options.install add checkbutton \
			-label "Shortcut icon to PgcEdit in Start Menu?" \
			-variable ::config(shortcut_startmenu) \
			-command {::winshellext::create_shortcut_to_PgcEdit startmenu $::config(shortcut_startmenu)}
	.mainmenu.options.install add checkbutton \
			-label "Shortcut icon to PgcEdit on Desktop?" \
			-variable ::config(shortcut_desktop) \
			-command {::winshellext::create_shortcut_to_PgcEdit desktop $::config(shortcut_desktop)}
	.mainmenu.options.install add checkbutton \
			-label "Shortcut icon to PgcEdit in Send To context menu?" \
			-variable ::config(shortcut_sendto) \
			-command {::winshellext::create_shortcut_to_PgcEdit sendto $::config(shortcut_sendto)}
	.mainmenu.options.install add command \
			-label "Create another shortcut icon to PgcEdit" \
			-command {::winshellext::create_shortcut_to_PgcEdit anydir true}
	.mainmenu.options.install add separator
	.mainmenu.options.install add checkbutton \
			-label "Associate PgcEdit with IFO files?" \
			-variable ::config(assoicated_with_IFO) \
			-command {::winshellext::associate_PgcEdit "IFO"}
	.mainmenu.options.install add checkbutton \
			-label "Associate PgcEdit with BUP files?" \
			-variable ::config(assoicated_with_BUP) \
			-command {::winshellext::associate_PgcEdit "BUP"}
	.mainmenu.options.install add separator
	.mainmenu.options.install add command \
			-label "Uninstall PgcEdit" \
			-command {::winshellext::uninstallPgcEdit}
}

.mainmenu.options add checkbutton -label "Launch Open DVD dialog at startup?" \
		-variable ::config(auto_open_new_dvd) -onvalue true -offvalue false
.mainmenu.options add checkbutton -label "Show PGC Selector at startup?" \
		-variable ::config(selector) -onvalue true -offvalue false
.mainmenu.options add checkbutton -label "Show Edit Buttons at startup?" \
		-variable ::config(buttons) -onvalue true -offvalue false
.mainmenu.options add checkbutton -label "Create backup when a DVD is opened for the first time?" \
		-variable ::config(auto_create_backup) -onvalue true -offvalue false
.mainmenu.options add checkbutton -label "Adjust Command Editor window size?" \
		-variable ::config(adjust_win_size) -onvalue true -offvalue false
.mainmenu.options add checkbutton -label "Tearoff menus? (require restart)" \
		-variable ::config(tearoff_menus) -onvalue true -offvalue false
.mainmenu.options add separator
.mainmenu.options add command -label "Global font (require restart)..." \
		-command {set f [::fontreq::fontreq .fr $::config(global_font) false]; \
				if {$f!={}} {set ::config(global_font) $f}}
.mainmenu.options add command -label "Fixed-width font..." \
		-command {set f [::fontreq::fontreq .fr $::config(fixed_width_font) true]; \
				if {$f!={}} {set ::config(fixed_width_font) $f; \
				.mf.f2.main.l configure -font $::config(fixed_width_font)}}
.mainmenu.options add separator
.mainmenu.options add checkbutton -label "Save setups on exit?" \
		-variable ::config(save_setups_on_exit) -onvalue true -offvalue false
.mainmenu.options add command -label "Save setups now" -command ::save_user_prefs
.mainmenu.options add command -label "Restore defaults (require restart)" \
		-command {file delete [file join $::env(HOME) ".pgcedit.cfg"]; \
				set ::config(restore_defaults) true}

.mainmenu add cascade -label "Help" -menu .mainmenu.help
menu .mainmenu.help
.mainmenu.help add command -label "About" -command ::main::about
.mainmenu.help add command -label "Doc (english)" -command {::main::showtextfile \
		.docen "PgcEdit Documentation" "$::config(apppath)/README.txt"}
.mainmenu.help add command -label "Doc (franais)" -command {::main::showtextfile \
		.docfr "PgcEdit Documentation" "$::config(apppath)/LISEZMOI.txt"}
.mainmenu.help add command -label "Versions history" -command {::main::showtextfile \
		.hist "PgcEdit Versions History" "$::config(apppath)/HISTORY.txt"}
.mainmenu.help add command -label "GPL License" -command {::main::showtextfile .license \
		"GNU General Public License" "$::config(apppath)/COPYING.txt"}

. configure -menu .mainmenu

# start is Scratchbook mode
scratchbook

# open selector if requested
if $::config(selector) {
	selector show ::main::labelslist ::main::cmdslist
}

wm title . "PgcEdit v $::program_version"
update
wm geometry . $::config(main_win_geometry)
wm deiconify .
update idletasks

# Get input filename from command line (if any)
if {$argc > 0} {
	set filename [lindex $argv 0]
	if [file isdirectory $filename] {
		Read_New_DVD [file normalize $filename]
	} else {
		Read_New_DVD [file normalize [file dirname $filename]]
	}
} else {
	if $::config(auto_open_new_dvd) Read_New_DVD
}

} ;# end namespace ::main
