#############################################################################
# streams.tcl Copyright 2004 r0lZ
# Compile informations about all Video/Audio/Sub-picture streams in the DVD.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::streams {
namespace export get_streams_info
namespace import ::readdvd::*

####### main proc
proc get_streams_info {} {
	if {! [info exists ::dvddir]} {return ""}

	set txt "DVD: \"$::dvddir/\""
	set txt "$txt\nDVD-TEXT General Name: \"[::utils::get_General_Name]\""
	set txt "$txt\nProvider ID: \"[::utils::get_provider_ID]\""
	set txt "$txt\nNumber of VTS: $::pgcs(numvts)"
	set txt "$txt\n"


	set numaudios [get_int ::raw_sectors(0,VMGM_MAT) 0x102 2]
	set numsubpics [get_int ::raw_sectors(0,VMGM_MAT) 0x154 2]
	set txt "$txt\nVMG Menu streams attributes (VIDEO_TS.IFO): $numaudios Audio, $numsubpics Sub-picture streams."
	set txt "$txt\n- Video: [video_attributes [get_int ::raw_sectors(0,VMGM_MAT) 0x100 2]]"
	for {set s 0} {$s < $numaudios} {incr s} {
		set offset [expr 0x104 + ( $s * 8 )]
		set attr [string range $::raw_sectors(0,VMGM_MAT) $offset [expr $offset + 7]]
		set txt "$txt\n- Audio $s: [audio_attributes $attr]"
	}
	if {$numsubpics > 0} {
		set attr [string range $::raw_sectors(0,VMGM_MAT) 0x156 0x15B]
		set txt "$txt\n- Subpic: [subpic_attributes $attr]"
	}

	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		set numaudios [get_int ::raw_sectors($vts,VTSI_MAT) 0x102 2]
		set numsubpics [get_int ::raw_sectors($vts,VTSI_MAT) 0x154 2]

		set txt "$txt\n"
		set txt "$txt\nVTS $vts Menu streams attributes (VTS_[format %02X $vts]_0.IFO): $numaudios Audio, $numsubpics Sub-picture streams."
		set txt "$txt\n- Video: [video_attributes [get_int ::raw_sectors($vts,VTSI_MAT) 0x100 2]]"
		for {set s 0} {$s < $numaudios} {incr s} {
			set offset [expr 0x104 + ( $s * 8 )]
			set attr [string range $::raw_sectors($vts,VTSI_MAT) $offset [expr $offset + 7]]
			set txt "$txt\n- Audio $s: [audio_attributes $attr]"
		}
		if {$numsubpics > 0} {
			set attr [string range $::raw_sectors($vts,VTSI_MAT) 0x156 0x15B]
			set txt "$txt\n- Subpic: [subpic_attributes $attr]"
		}

		set numaudios [get_int ::raw_sectors($vts,VTSI_MAT) 0x202 2]
		set numsubpics [get_int ::raw_sectors($vts,VTSI_MAT) 0x254 2]
		set txt "$txt\n"
		set txt "$txt\nVTS $vts VOB streams attributes (VTS_[format %02X $vts]_0.IFO): $numaudios Audio, $numsubpics Sub-picture streams."
		set txt "$txt\n- Video: [video_attributes [get_int ::raw_sectors($vts,VTSI_MAT) 0x200 2]]"
		for {set s 0} {$s < $numaudios} {incr s} {
			set offset [expr 0x204 + ( $s * 8 )]
			set attr [string range $::raw_sectors($vts,VTSI_MAT) $offset [expr $offset + 7]]
			set txt "$txt\n- Audio $s: [audio_attributes $attr]"
		}
		for {set s 0} {$s < $numsubpics} {incr s} {
			set offset [expr 0x256 + ( $s * 6 )]
			set attr [string range $::raw_sectors($vts,VTSI_MAT) $offset [expr $offset + 5]]
			set txt "$txt\n- Subpic $s: [subpic_attributes $attr]"
		}
	}
	::main::showtext .streamsinfo "PgcEdit: Streams information" $txt
	return $txt
}


proc video_attributes {vattr} {

	if {[expr $vattr & 0x4000] == 0} {
		set coding_mode "MPEG-1"
	} else {
		set coding_mode "MPEG-2"
	}

	if {[expr $vattr & 0x1000] == 0} {
		set standard "NTSC"
	} else {
		set standard "PAL"
	}

	if {[expr $vattr & 0x0C00] == 0} {
		set aspect "4:3"
	} else {
		set aspect "16:9"
	}

	if {[expr $vattr & 0x0010] == 0} {
		set bitrate "VBR"
	} else {
		set bitrate "CBR"
	}

	set m [expr ($vattr & 0x0300) / 0x100]
	switch $m {
		3 {set auto_mode ""}
		2 {set auto_mode ", Automatic Letterbox"}
		1 {set auto_mode ", Automatic Pan&Scan"}
		0 {set auto_mode ", Automatic Letterbox & Pan&Scan"}
		default {set auto_mode ", Unknow Automatic mode #$m"}
	}

	set r [expr ($vattr & 0x000C) / 4]
	if {$standard == "NTSC"} {
		set h 480
	} else {
		set h 576
	}
	switch $r {
		0 {set w 720}
		1 {set w 704}
		2 {set w 352}
		3 {set w 352; set h [expr $h / 2]}
	}
	set resolution ${w}x${h}

	if {[expr $vattr & 0x0002] == 0} {
		set letterboxed "not cropped"
	} else {
		set letterboxed "cropped"
	}

	return "$standard $aspect, $resolution, $letterboxed ($coding_mode $bitrate)$auto_mode"
}

proc audio_attributes {aattr} {

	set byte0 [get_int aattr 0 1]

	set c [expr ($byte0 & 0xE0) / 0x20]
	switch $c {
		0 {set coding_mode "AC3"}
		2 {set coding_mode "MPEG-1"}
		3 {set coding_mode "MPEG-2"}
		4 {set coding_mode "LPCM"}
		6 {set coding_mode "DTS"}
		default {set coding_mode "Unknow coding mode #$c"}
	}

	if {[expr $byte0 & 0x0C] == 0} {
		set lang_code "--"
	} else {
		set lang_code [string range $aattr 2 3]
	}

	set a [expr $byte0 & 0x03]
	switch $a {
		0 {set applic_mode "Unspecified application mode"}
		1 {set applic_mode "Karaoke"}
		2 {
			set byte7 [get_int aattr 7 1]
			if {[expr $byte7 & 8] == 8} {
				set applic_mode "Dolby-Surround"
			} else {
				set applic_mode "Surround"
			}
		}
		3 {set applic_mode "Unknow application mode #$a"}
	}

	set byte1 [get_int aattr 1 1]

	set num_channels [expr ($byte1 & 7) + 1]

	set byte5 [get_int aattr 5 1]

	switch $byte5 {
		0 {set code_ext "unspecified"}
		1 {set code_ext "normal"}
		2 {set code_ext "for visually impaired"}
		3 {set code_ext "director's comments"}
		4 {set code_ext "alternate director's comments"}
		default {set code_ext "unknow extension #$byte5"}
	}

	return "$lang_code, $code_ext, $coding_mode ($num_channels channels), $applic_mode"
}

proc subpic_attributes {spattr} {

	set byte0 [get_int spattr 0 1]

	if {[expr $byte0 & 0x03] == 0} {
		set lang_code "--"
	} else {
		set lang_code [string range $spattr 2 3]
	}

	set byte5 [get_int spattr 0 1]

	switch $byte5 {
		0 {set code_ext "unspecified extension"}
		1 {set code_ext "normal"}
		2 {set code_ext "large"}
		3 {set code_ext "children"}
		5 {set code_ext "normal captions"}
		6 {set code_ext "large captions"}
		7 {set code_ext "children captions"}
		9 {set code_ext "forced"}
		13 {set code_ext "director comments"}
		14 {set code_ext "large director comments"}
		15 {set code_ext "director comments for children"}
		default {set code_ext "unknow extension #$byte5"}
	}

	return "$lang_code, $code_ext"
}

}	;# end namespace streams
 