#############################################################################
# utils.tcl Copyright 2004 r0lZ
# Procedures related to the Utilities menu.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::utils {
namespace export utils_menu
namespace import ::readdvd::*

#############################################################################
## utils_menu
## build the menu with all utils functions defined in this file
proc utils_menu {menuname} {
	menu $menuname
	$menuname add command -label "Title Play Map table..." -command ::srptinfo::srptinfo
	$menuname add command -label "Streams information..." -command ::streams::get_streams_info
	$menuname add separator
	$menuname add command -label "Filter Prohibited User Operations..." -command {::utils::filter_puos}
	$menuname add separator
	$menuname add command -label "Kill PGC playback..." -command {::utils::kill_pgc_playback}
	$menuname add separator
	$menuname add command -label "Find Unused GPRMs..." -command {::utils::find_unused_gprms}
	$menuname add separator
	$menuname add command -label "New Menu..." -command {::utils::new_menu}
	$menuname add command -label "Import Menu..." -command {::utils::import_menu}
	$menuname add command -label "Remove Menu..." -command {::utils::remove_menu}
	$menuname add separator
	$menuname add command -label "New PGC..." -command ::utils::create_new_pgc
	$menuname add separator
	$menuname add command -label "Set Provider ID..." -command ::utils::set_provider_ID
	$menuname add command -label "Set DVD-TEXT General Name..." -command ::utils::set_General_Name
	$menuname add command -label "Remove DVD-TEXT data..." -command ::utils::remove_dvd_text_data
	$menuname add separator
	$menuname add command -label "DVDShrink Title Button..." -command {::utils::title_but_next_title}
}

#############################################################################
## utility procedures

# return the number of pre, post or cell commands in the provided commands
proc num_type_cmds {type cmds_upvar} {
	upvar $cmds_upvar cmds

	set idxs [lsearch -all $cmds {[*]*}]
	switch $type {
		pre  {return [expr [lindex $idxs 1]-[lindex $idxs 0]-1]}
		post {return [expr [lindex $idxs 2]-[lindex $idxs 1]-1]}
		cell {return [expr [llength $cmds]-[lindex $idxs 2]-1]}
	}
}

# return the pre, post or cell commands from the provided commands
proc get_type_cmds {type cmds_upvar} {
	upvar $cmds_upvar cmds

	set idxs [lsearch -all $cmds {[*]*}]
	switch $type {
		pre  {return [lrange $cmds [expr [lindex $idxs 0]+1] [expr [lindex $idxs 1]-1]]}
		post {return [lrange $cmds [expr [lindex $idxs 1]+1] [expr [lindex $idxs 2]-1]]}
		cell {return [lrange $cmds [expr [lindex $idxs 2]+1] end]}
	}
}

# replace the pre, post or cell commands in the commands list by the provided commands
proc set_type_cmds {type cmds_upvar newcmds} {
	upvar $cmds_upvar cmds

	set idxs [lsearch -all $cmds {[*]*}]
	switch $type {
		pre  {
			set idx1 [lindex $idxs 0]
			set idx2 [lindex $idxs 1]
		}
		post {
			set idx1 [lindex $idxs 1]
			set idx2 [lindex $idxs 2]
		}
		cell {
			set idx1 [lindex $idxs 2]
			set idx2 [llength $cmds]
		}
	}
	set cmds [concat [lrange $cmds 0 $idx1] $newcmds [lrange $cmds $idx2 end]]
}

# refresh the listbox with the current PGC
proc refresh_current_pgc {} {
	if $::scratchbook_active {
		return
	}
	set selectionindexes [::main::get_selection_indexes]
	set type [lindex $::editor_cmds_from 0]
	set pgcindex [split [lindex $::editor_cmds_from 1] ","]
	set vts [lindex $pgcindex 0]
	switch $type {
		pgc {
			::main::display_pgc $type $vts [lindex $pgcindex 1]
		}
		menu {
			::main::display_pgc $type $vts [lindex $pgcindex 1] [lindex $pgcindex 2]
		}
	}
	::main::set_listbox_selection [lindex $selectionindexes 0] [lindex $selectionindexes 1]
}

#############################################################################
## Filter prohibited user operations
proc filter_puos {} {

	### sanity check
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Filter Prohibited User Operations"

	set txt "This utility will selectively clear the PUOs"
	set txt "$txt\nin all PGCs of the VMG Menus,"
	set txt "$txt\nin all PGCs of the VTS Menus,"
	set txt "$txt\nin all PGCs of the VTS Titles,"
	set txt "$txt\nand in the VMG Title Play Map Table."
	set txt "$txt\n"
	set txt "$txt\nContinue ?"
	set r [tk_messageBox -title $utiltitle -icon question -message $txt \
			-parent . -type yesno -default no]
	if {$r == "no"} return

	# VMG Menus PUOs
	set puos 0
	for {set lu 1} {$lu <= $::menus(0,numlus)} {incr lu} {
		for {set pgc 1} {$pgc <= $::menus(0,$lu,numpgcs)} {incr pgc} {
			set puos [expr $puos | [get_int ::menus(0,$lu,$pgc,header) 8 4]]
		}
	}
	if {$puos == 0} {
		tk_messageBox -title $utiltitle -icon info -parent . -type ok \
				-message "No Prohibited User Operations found in the VMG Menus."
	} else {
		set ::pgced::puos $puos
		set filterpuos [::pgced::puos_editor "Filter PUOs for all VMG Menus :\nClear the PUOs you want to clear,\nSet the PUOs you don't want to change." .]
		if {$filterpuos != $puos} {
			::main::dvd_modified true false
			for {set lu 1} {$lu <= $::menus(0,numlus)} {incr lu} {
				for {set pgc 1} {$pgc <= $::menus(0,$lu,numpgcs)} {incr pgc} {
					::writedvd::set_int ::menus(0,$lu,$pgc,header) 8 \
							[expr $filterpuos & [get_int ::menus(0,$lu,$pgc,header) 8 4]] 4
				}
			}
		}
	}

	# VTS Menus PUOs
	set puos 0
	for {set vts 1} {$vts <= $::menus(numvts)} {incr vts} {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set puos [expr $puos | [get_int ::menus($vts,$lu,$pgc,header) 8 4]]
			}
		}
	}
	if {$puos == 0} {
		tk_messageBox -title $utiltitle -icon info -parent . -type ok \
				-message "No Prohibited User Operations found in the VTS Menus."
	} else {
		set ::pgced::puos $puos
		set filterpuos [::pgced::puos_editor "Filter PUOs for all VTS Menus :\nClear the PUOs you want to clear,\nSet the PUOs you don't want to change." .]
		if {$filterpuos != $puos} {
			::main::dvd_modified true false
			for {set vts 1} {$vts <= $::menus(numvts)} {incr vts} {
				for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
					for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
						::writedvd::set_int ::menus($vts,$lu,$pgc,header) 8 \
								[expr $filterpuos & [get_int ::menus($vts,$lu,$pgc,header) 8 4]] 4
					}
				}
			}
		}
	}

	# Titles PUOs
	set puos 0
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set ttn 1} {$ttn <= $::pgcs($vts,numpgcs)} {incr ttn} {
			set puos [expr $puos | [get_int ::pgcs($vts,$ttn,header) 8 4]]
		}
	}
	if {$puos == 0} {
		tk_messageBox -title $utiltitle -icon info -parent . -type ok \
				-message "No Prohibited User Operations found in the Title's PGCs."
	} else {
		set ::pgced::puos $puos
		set filterpuos [::pgced::puos_editor "Filter PUOs for all Titles :\nClear the PUOs you want to clear,\nSet the PUOs you don't want to change." .]
		if {$filterpuos != $puos} {
			::main::dvd_modified true false
			for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
				for {set ttn 1} {$ttn <= $::pgcs($vts,numpgcs)} {incr ttn} {
					::writedvd::set_int ::pgcs($vts,$ttn,header) 8 \
							[expr $filterpuos & [get_int ::pgcs($vts,$ttn,header) 8 4]] 4
				}
			}
		}
	}

	# VMG_PTT_SRPT PUOs
	set numtitles [get_int ::raw_sectors(0,VMG_PTT_SRPT) 0 2]
	set puos 0
	for {set t 1} {$t <= $numtitles} {incr t} {
		set t_idx [expr 8+(($t-1)*12)]
		set puos [expr $puos | ([get_int ::raw_sectors(0,VMG_PTT_SRPT) $t_idx 1] & 3)]
	}
	if {$puos == 0} {
		tk_messageBox -title $utiltitle -icon info -parent . -type ok \
				-message "No Prohibited User Operations found in the Title Play Map Table."
	} else {
		set filterpuos $puos
		if {[expr $puos & 1]} {
			set txt "At least one Title has the \"Time Play or Search\" prohibited"
			set txt "$txt\nin the VMG Title Play Map Table."
			set txt "$txt\n"
			set txt "$txt\nClear this prohibition for all Titles ?"
			set r [tk_messageBox -title $utiltitle -icon question -message $txt \
					-parent . -type yesno -default no]
			if {$r == "yes"} {set filterpuos [expr $filterpuos & 2]}
		}
		if {[expr $puos & 2]} {
			set txt "At least one Title has the \"PTT Play or Search\" prohibited"
			set txt "$txt\nin the VMG Title Play Map Table."
			set txt "$txt\n"
			set txt "$txt\nClear this prohibition for all Titles ?"
			set r [tk_messageBox -title $utiltitle -icon question -message $txt \
					-parent . -type yesno -default no]
			if {$r == "yes"} {set filterpuos [expr $filterpuos & 1]}
		}
		if {$puos != $filterpuos} {
			::main::dvd_modified true false
			set mask [expr $filterpuos | 0xFC]
			for {set t 1} {$t <= $numtitles} {incr t} {
				set t_idx [expr 8+(($t-1)*12)]
				set byte [expr [get_int ::raw_sectors(0,VMG_PTT_SRPT) $t_idx 1] & $mask]
				::writedvd::set_int ::raw_sectors(0,VMG_PTT_SRPT) $t_idx $byte 1
			}
		}
	}
}

#############################################################################
## New (create or replace) the VMGM_PGCI_UT Title Menu PGC
proc new_title_menu {} {

	### sanity check
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "New VMG Title Menu PGC"
	set txt "This utility will create the VMG Title menu PGC."
	set txt "$txt\nNo VOB will be associated with the menu."
	set txt "$txt\n"
	set txt "$txt\nContinue ?"
	set r [tk_messageBox -title $utiltitle -icon question -message $txt \
			-parent . -type yesno -default no]
	if {$r == "no"} return

	### create or replace the VMG menu
	if {$::menus(0,numlus)>0} {
		set txt "WARNING: There is already a VMG Title Menu!"
		set txt "$txt\nIf you continue, it will be REPLACED by the new one!"
		set txt "$txt\n"
		set txt "$txt\nContinue anyway ?"
		set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	}
	set ::raw_sectors(0,VMGM_PGCI_UT) \
			[readFile $::config(apppath)/lib/data/vmgm_pgci_ut_wo_vob.sec binary]

	parse_PGCI_UT ::raw_sectors(0,VMGM_PGCI_UT) 0 ::menus
	::main::init_tree_menu
	::main::dvd_modified true false
	::main::display_pgc menu 0 1 1

	set vobname "$::dvddir/VIDEO_TS.VOB"
	if {[file exists $vobname]} {
		set txt "There is an old $vobname file."
		set txt "$txt\n\nDelete it ?"
		set r [tk_messageBox -title $utiltitle -icon question -message $txt \
				-parent . -type yesno -default yes]
		if {$r == "yes"} {
			file delete $vobname
		}
	}
}

#############################################################################
## New (create or replace) a VTSM_PGCI_UT Root Menu PGC
proc new_root_menu {vtsnum} {

	### sanity check
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	set utiltitle "New VTS Root Menu PGC"
	set txt "This utility will create the VTS $vtsnum Root menu PGC."
	set txt "$txt\nNo VOB will be associated with the menu."
	set txt "$txt\n"
	set txt "$txt\nContinue ?"
	set r [tk_messageBox -title $utiltitle -icon question -message $txt \
			-parent . -type yesno -default no]
	if {$r == "no"} return

	### create or replace the VTS menu
	if {$::menus($vtsnum,numlus)>0} {
		set txt "WARNING: There is already a menu for VTS $vtsnum!"
		set txt "$txt\nIf you continue, it will be REPLACED by the new one!"
		set txt "$txt\n"
		set txt "$txt\nContinue anyway ?"
		set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	}

	set ::raw_sectors($vtsnum,VTSM_PGCI_UT) \
			[readFile $::config(apppath)/lib/data/vtsm_pgci_ut_wo_vob.sec binary]

	parse_PGCI_UT ::raw_sectors($vtsnum,VTSM_PGCI_UT) $vtsnum ::menus
	::main::init_tree_menu
	::main::dvd_modified true false
	::main::display_pgc menu $vtsnum 1 1

	set vobname "$::dvddir/VTS_[format %02d $vtsnum]_0.VOB"
	if {[file exists $vobname]} {
		set txt "There is an old $vobname file."
		set txt "$txt\n\nDelete it ?"
		set r [tk_messageBox -title $utiltitle -icon question -message $txt \
				-parent . -type yesno -default yes]
		if {$r == "yes"} {
			file delete $vobname
		}
	}
}

proc new_menu {} {
	### sanity check
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set vtsnum [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
	if {$vtsnum == 0} {
		new_title_menu
	} else {
		new_root_menu $vtsnum
	}
}

proc import_menu {} {
	### sanity check
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Import Menu"
	set vtsnum [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
	if {$vtsnum == 0} {
		set current "VMG"
	} else {
		set current "VTS $vtsnum"
	}

	if {$::menus($vtsnum,numlus)>0} {
		set txt "WARNING: There is already a menu for the $current!"
		set txt "$txt\nIf you continue, it will be REPLACED by the new one!"
		set txt "$txt\n"
		set txt "$txt\nContinue anyway ?"
		set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	}

	set sourceifofile [tk_getOpenFile -initialdir $::config(importmenuinitialdir) \
		-filetypes {{"IFO files" {*.IFO}}} \
		-title "Select the IFO file with the menu to import"]
	if {$sourceifofile == ""} {return}
	set sourceifofile [file nativename $sourceifofile]
	set ::config(importmenuinitialdir) [file dirname $sourceifofile]

	set sourcevtsnum [readdvdfile $sourceifofile ::import_menu_sectors ::import_menu_pgcs ::import_menu_menus]
	if {$sourcevtsnum == -1} return
	if {$sourcevtsnum == 0} {
		set sourcetype "VMG"
	} else {
		set sourcetype "VTS"
	}

	if {$::import_menu_sectors($sourcevtsnum,${sourcetype}M_PGCI_UT) == ""} {
		set r [tk_messageBox -title $utiltitle -icon info -parent . -type ok \
				-message "Sorry, there is no menu defined in file\n\"$sourceifofile\""]
		return
	}
	if {$vtsnum == 0} {
		set destvtstype "VMG"
		set destvobfilename [file join $::dvddir "VIDEO_TS.VOB"]
	} else {
		set destvtstype "VTS"
		set destvobfilename [file join $::dvddir "VTS_[format %02i $vtsnum]_0.VOB"]
	}
	set ::raw_sectors($vtsnum,${destvtstype}M_PGCI_UT) \
			$::import_menu_sectors($sourcevtsnum,${sourcetype}M_PGCI_UT)
	set ::raw_sectors($vtsnum,${destvtstype}M_C_ADT) \
			$::import_menu_sectors($sourcevtsnum,${sourcetype}M_C_ADT)
	set ::raw_sectors($vtsnum,${destvtstype}M_VOBU_ADMAP) \
			$::import_menu_sectors($sourcevtsnum,${sourcetype}M_VOBU_ADMAP)

	parse_PGCI_UT ::raw_sectors($vtsnum,${destvtstype}M_PGCI_UT) $vtsnum ::menus
	::main::init_tree_menu
	::main::display_pgc menu $vtsnum 1 1
	::main::dvd_modified true false

	if {$::import_menu_sectors($sourcevtsnum,${sourcetype}M_VOBU_ADMAP) != ""} {
		set sourcevobfilename "[file rootname $sourceifofile].VOB"
		if [file exists $sourcevobfilename] {
			set txt "Copy the Menu VOB file ?"
			if [file exists $destvobfilename] {
				set txt "$txt\n\nWARNING: Current VOB file will be overwrited!"
			}
			set r [tk_messageBox -title $utiltitle -icon question -message $txt \
					-parent . -type yesno -default yes]
			if {$r == "yes"} {
				file copy -force $sourcevobfilename $destvobfilename
			}
		}
	}

	unset ::import_menu_sectors ::import_menu_pgcs ::import_menu_menus
}

proc remove_menu {} {
	### sanity check
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Remove Menu"
	set vtsnum [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
	if {$vtsnum == 0} {
		set current "VMG"
		set vtstype "VMG"
		set vobfilename [file join $::dvddir "VIDEO_TS.VOB"]
	} else {
		set current "VTS $vtsnum"
		set vtstype "VTS"
		set vobfilename [file join $::dvddir "VTS_[format %02i $vtsnum]_0.VOB"]
	}

	if {$::menus($vtsnum,numlus)>0} {
		set txt "This utility will remove the whole Menu from the $current."
		set txt "$txt\n"
		set txt "$txt\nAre you sure ?"
		set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	} else {
		tk_messageBox -title $utiltitle -icon error -parent . -type ok \
				-message "There is no menu for the $current!"
		return
	}

	if [file exists $vobfilename] {
		set r [tk_messageBox -title $utiltitle -icon question -parent . -type yesno -default yes \
				-message "Delete the Menu VOB file ?"]
		if {$r == "yes"} {file delete $vobfilename}
	}

	set ::raw_sectors($vtsnum,${vtstype}M_PGCI_UT) ""
	set ::raw_sectors($vtsnum,${vtstype}M_C_ADT) ""
	set ::raw_sectors($vtsnum,${vtstype}M_VOBU_ADMAP) ""
	set ::menus($vtsnum,numlus) 0

	::main::init_tree_menu
	::main::dvd_modified true false
	if {[lindex $::editor_cmds_from 0] == "menu"} {
		::main::display_pgc pgc 0 0
	}
}

proc create_new_pgc {} {
	### sanity check
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set pgctype [lindex $::editor_cmds_from 0]
	set vtsnum [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
	set cellpbtime0 [expr [get_int ::pgcs(0,0,header) 4 4] & 0xC0]	;# NTSC or PAL 0 length time

	set utiltitle "Create New PGC"
	set txt "This utility will create a new, blank PGC,"
	set txt "$txt\nat the end of the current PGC group"
	set txt "$txt\n(end of Menu-LU or TTNs)."
	set txt "$txt\n"
	set txt "$txt\nShould I create the following PGC?"
	if {$pgctype == "menu"} {
		set lunum [lindex [split [lindex $::editor_cmds_from 1] ","] 1]
		set numpgcs $::menus($vtsnum,$lunum,numpgcs)
		set pgcnum [expr $numpgcs + 1]
		if {$vtsnum == 0} {
			set txt "$txt\nVMG"
		} else {
			set txt "$txt\nVTS $vtsnum"
		}
		set txt "$txt / Menu-LU $lunum ($::menus($vtsnum,$lunum,lang)) / $pgcnum"
	} else {
		if {$vtsnum == 0} {
			set txt "Can't add a new PGC to the First Play PGC!"
			set txt "$txt\nPlease select another PGC."
			tk_messageBox -title $utiltitle -icon error -message $txt \
					-parent . -type ok
			return
		}
		set numpgcs $::pgcs($vtsnum,numpgcs)
		set pgcnum [expr $numpgcs + 1]
		set txt "$txt\nVTS $vtsnum / TTN $pgcnum"
	}
	set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
			-parent . -type yesno -default yes]
	if {$r == "no"} return

	if {$pgctype == "menu"} {
		set ::menus($vtsnum,$lunum,$pgcnum,mask) 0
		set ::menus($vtsnum,$lunum,$pgcnum,masktext) ""
		set ::menus($vtsnum,$lunum,$pgcnum) {{* pre} {00 00 00 00 00 00 00 00} {* post} {* cell}}
		set ::menus($vtsnum,$lunum,$pgcnum,header) [string repeat \000 236]
		::writedvd::set_int ::menus($vtsnum,$lunum,$pgcnum,header) 4 $cellpbtime0 4
		set ::menus($vtsnum,$lunum,$pgcnum,numprograms) 0
		set ::menus($vtsnum,$lunum,$pgcnum,programmap) ""
		set ::menus($vtsnum,$lunum,$pgcnum,numcells) 0
		set ::menus($vtsnum,$lunum,$pgcnum,cellpbtable) ""
		set ::menus($vtsnum,$lunum,$pgcnum,cellpostable) ""
		if {$vtsnum == 0} {
			set ::raw_sectors($vtsnum,VMGM_PGCI_UT) "$::raw_sectors($vtsnum,VMGM_PGCI_UT)[string repeat \000 8]"
		} else {
			set ::raw_sectors($vtsnum,VTSM_PGCI_UT) "$::raw_sectors($vtsnum,VTSM_PGCI_UT)[string repeat \000 8]"
		}
		incr ::menus($vtsnum,$lunum,numpgcs)

		::main::init_tree_menu
		::main::display_pgc menu $vtsnum $lunum $pgcnum
	} else {
		set ::pgcs($vtsnum,$pgcnum,mask) 0
		set ::pgcs($vtsnum,$pgcnum,masktext) ""
		set ::pgcs($vtsnum,$pgcnum) {{* pre} {00 00 00 00 00 00 00 00} {* post} {* cell}}
		set ::pgcs($vtsnum,$pgcnum,header) [string repeat \000 236]
		::writedvd::set_int ::pgcs($vtsnum,$pgcnum,header) 4 $cellpbtime0 4
		set ::pgcs($vtsnum,$pgcnum,numprograms) 0
		set ::pgcs($vtsnum,$pgcnum,programmap) ""
		set ::pgcs($vtsnum,$pgcnum,numcells) 0
		set ::pgcs($vtsnum,$pgcnum,cellpbtable) ""
		set ::pgcs($vtsnum,$pgcnum,cellpostable) ""
		set ::raw_sectors($vtsnum,VTS_PGCI) "$::raw_sectors($vtsnum,VTS_PGCI)[string repeat \000 8]"
		incr ::pgcs($vtsnum,numpgcs)

		::main::init_tree_menu
		::main::display_pgc pgc $vtsnum $pgcnum
	}
	::main::dvd_modified true false
	return
}

#############################################################################
## add title button to skip to the next title on a DVD reauthored with DVDShrink
proc title_but_next_title {} {

	### sanity check
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Title button to skip to next title for DVDShrink"
	set txt "This utility work ONLY on DVD files produced by"
	set txt "$txt\nDVDShrink in reauthor mode, with multiple Titles!"
	set txt "$txt\n"
	set txt "$txt\nThis utility will modify the commands of (almost) every"
	set txt "$txt\nPGCs, and create the VMG Title menu PGC, so that the"
	set txt "$txt\nuser can use the Title button of the remote to skip to"
	set txt "$txt\nthe next title."
	set txt "$txt\n"
	set txt "$txt\nThanks goes to 2COOL for the idea.  See this tread:"
	set txt "$txt\nhttp://forum.doom9.org/showthread.php?threadid=61238"
	set txt "$txt\n"
	set txt "$txt\nWARNING: Only some basic checks are made to ensure"
	set txt "$txt\nthat the DVD is in the correct format."
	set txt "$txt\nUse this utility only with unmodified DVDShrink"
	set txt "$txt\nreauthored DVDs!"
	set txt "$txt\nALWAYS backup the DVD IFOs before using this utility!"
	set txt "$txt\n"
	set txt "$txt\nContinue ?"
	set r [tk_messageBox -title $utiltitle -icon question -message $txt \
			-parent . -type yesno -default no]
	if {$r == "no"} return
	if {$::pgcs(numvts)<=1} {
		set txt "WARNING: There is only $::pgcs(numvts) VTS in this DVD!"
		set txt "$txt\nApplying this utility is therefore useless!"
		set txt "$txt\n"
		set txt "$txt\nContinue anyway ?"
		set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	}

	### create or replace the VMG menu
	if {$::menus(0,numlus)>0} {
		set txt "WARNING: There is already a VMG Title Menu!"
		set txt "$txt\nIf you continue, it will be REPLACED by a new one!"
		if {[file exists "$::dvddir/VIDEO_TS.VOB"]} {
			set txt "$txt\n"
			set txt "$txt\nFurthermore, there is a VIDEO_TS.VOB file."
			set txt "$txt\nYou will need to manually delete it!"
		}
		set txt "$txt\n"
		set txt "$txt\nContinue anyway ?"
		set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	}
	set ::raw_sectors(0,VMGM_PGCI_UT) \
			[readFile $::config(apppath)/lib/data/vmgm_pgci_ut_wo_vob.sec binary]

	parse_PGCI_UT ::raw_sectors(0,VMGM_PGCI_UT) 0 ::menus
	::main::init_tree_menu
	::main::dvd_modified true false

	### for each VTS except last one, replace the pre command "Set gprm(0) = 0"
	### by "Set gprm(0) = <current VTS number>"
	for {set vts 1} {$vts < $::pgcs(numvts)} {incr vts} {
		set cmds [get_type_cmds pre ::pgcs($vts,1)]
		set idx [lsearch -regexp $cmds {71 00 00 00 .. .. 00 00}]
		if {$idx == -1} {
			set cmds [linsert $cmds 0 "71 00 00 00 00 [format %02X $vts] 00 00"]
		} else {
			lset cmds $idx "71 00 00 00 00 [format %02X $vts] 00 00"
		}
		set_type_cmds pre ::pgcs($vts,1) $cmds
	}

	### Last VTS: modify the last post command (normally "exit") to
	### "{CallSS-FP} Call the First Play PGC, resume cell 1" or "exit"
	set vts $::pgcs(numvts)
	set cmds [get_type_cmds post ::pgcs($vts,1)]
	set idx [lsearch $cmds {30 01 00 00 00 00 00 00}]
	if {$idx == -1} {
		set idx [lsearch $cmds {30 08 00 00 01 00 00 00}]
	}
	if {$idx >= 0} {
		set txt "Loop option:"
		set txt "$txt\n"
		set txt "$txt\nDo you want the DVD to loop back"
		set txt "$txt\nto the first title at the end ?"
		set r [tk_messageBox -title "Title button to skip to next title for DVDShrink" \
				-icon question -message $txt -parent . -type yesno -default no]
		if {$r == "yes"} {
			lset cmds $idx "30 08 00 00 01 00 00 00"
		} else {
			lset cmds $idx "30 01 00 00 00 00 00 00"
		}
	}
	set_type_cmds post ::pgcs($vts,1) $cmds

	### util done.  Refresh listbox.
	refresh_current_pgc
}


#############################################################################
## Procedure:  set_provider_ID
## Show/Change the provider ID
proc {set_provider_ID} {} {
	if [info exists ::raw_sectors(0,VMGM_MAT)] {
		# Get old Provider ID (if any)
		set oldid [get_provider_ID]
		# get new ID
		set newid [::stringreq::stringreq -window .getid -def $oldid -ok "Set ID" \
				-callback ::utils::set_provider_ID_callback -width 32 \
				-text "Enter the DVD Provider ID:" -title "PgcEdit" -geometry w]
		if {$newid != ""} {
			set newid "$newid[string repeat \000 [expr 32-[string length $newid]]]"
			set ::raw_sectors(0,VMGM_MAT) \
					[string replace $::raw_sectors(0,VMGM_MAT) 0x40 0x5F $newid]
			::main::set_window_title
			::main::dvd_modified true false
		}
	} else {
		bell
	}
}
proc {set_provider_ID_callback} {str} {
	if {[string length $str] > 32} {
		bell
		return 0
	} else {
		return 1
	}
}
proc {get_provider_ID} {} {
	if [info exists ::raw_sectors(0,VMGM_MAT)] {
		set id [string range $::raw_sectors(0,VMGM_MAT) 0x40 0x5F]
		set stopchar [string first \000 $id]
		if {$stopchar != -1} {
			set id [string range $id 0 [expr $stopchar - 1]]
		}
		return [string trim $id " "]
	} else {
		return ""
	}
}

#############################################################################
## Procedure:  set_General_Name
## Show/Change the provider ID
proc {set_General_Name} {} {
	if {![info exists ::dvddir]} {
		bell
		return
	}
	if {$::raw_sectors(0,VMG_TXTDT_MG) != ""} {
		set newsec false
		set newmsg "VMG_TXTDT_MG (DVD-TEXT data sector) already exists."
		## some sanity check (should be more accurate!)
		if {[string range $::raw_sectors(0,VMG_TXTDT_MG) 0 8] != "TEXT_DATA"} {
			bell
			return
		}
		if {[get_int ::raw_sectors(0,VMG_TXTDT_MG) 0xF0 4] != [expr 0x3000000C]} {
			bell
			return
		}
	} else {
		set newsec true
		set newmsg "The VMG_TXTDT_MG (DVD-TEXT data sector) will be created."
		set ::raw_sectors(0,VMG_TXTDT_MG) \
				[readFile $::config(apppath)/lib/data/vmg_txtdt_mg_default.sec binary]
	}

	# Get old Provider ID (if any)
	set oldid [get_General_Name]
	# get new ID
	set newid [::stringreq::stringreq -window .getid -def $oldid -ok "Set Name" \
			-callback ::utils::set_general_name_callback -width 30 \
			-text "$newmsg\n\nEnter the DVD-TEXT General Name:" -title "PgcEdit" -geometry w]
	if {$newid != ""} {
		set newid "$newid[string repeat \000 [expr 30-[string length $newid]]]"
		set ::raw_sectors(0,VMG_TXTDT_MG) \
				[string replace $::raw_sectors(0,VMG_TXTDT_MG) 0xF4 0x111 $newid]
		::main::set_window_title
		::main::dvd_modified true false
	} elseif $newsec {
		set ::raw_sectors(0,VMG_TXTDT_MG) ""
	}
}
proc {set_general_name_callback} {str} {
	if {[string length $str] > 30} {
		bell
		return 0
	} else {
		return 1
	}
}
proc {get_General_Name} {} {
	if {![info exists ::dvddir]} {
		bell
		return ""
	}
	set name ""
	if {$::raw_sectors(0,VMG_TXTDT_MG) != ""} {
		set nameid [get_int ::raw_sectors(0,VMG_TXTDT_MG) 0xF0 4]
		if {$nameid == [expr 0x3000000C]} {
			set name [string range $::raw_sectors(0,VMG_TXTDT_MG) 0xF4 0x111]
			set stopchar [string first \000 $name]
			if {$stopchar != -1} {
				set name [string range $name 0 [expr $stopchar - 1]]
			}
			set name [string trim $name " "]
		}
	}
	return $name
}
proc {remove_dvd_text_data} {} {
	if {![info exists ::dvddir]} {
		bell
		return
	}
	if {$::raw_sectors(0,VMG_TXTDT_MG) != ""} {
		set msg "This will remove the whole VMG_TXTDT_MG\n(DVD-TEXT data sector), not only the\nGeneral Name!\n\nAre you sure ?"
		set r [tk_messageBox -title "Remove DVD-TEXT data" -icon question -message $msg \
				-parent . -type okcancel]
		if {$r == "ok"} {
			set ::raw_sectors(0,VMG_TXTDT_MG) ""
			::main::set_window_title
			::main::dvd_modified true false
		}
	} else {
		tk_messageBox -title "Remove DVD-TEXT data" -icon error -parent . -type ok \
				-message "There is no VMG_TXTDT_MG (DVD-TEXT\ndata sector) in this DVD."
	}
}

proc kill_pgc_playback {} {
	### sanity check
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	set txt "This utility will inhibit the video playback"
	set txt "$txt\nof the current PGC."
	set txt "$txt\nThe VOB file will not be modified."
	set txt "$txt\n"
	set txt "$txt\nIf you want to revert to the original"
	set txt "$txt\nPGC behaviour, simply remove the last"
	set txt "$txt\npre-command (LinkTailPGC)."
	set txt "$txt\n"
	set txt "$txt\nContinue ?"
	set r [tk_messageBox -title "Kill PGC Playback" -icon warning -message $txt \
			-parent . -type yesno -default yes]
	if {$r == "no"} return

	# get pre-commands, and add a NOP and a LinkTailPGC commands at the end.
	set pgcvar ::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1])
	set cmds [get_type_cmds pre $pgcvar]
	lappend cmds {00 00 00 00 00 00 00 00}	;# add NOP
	set gotocmd [llength $cmds]	;# we will change the Breaks to Gotos to this line number
	lappend cmds {20 01 00 00 00 00 00 0D}	;# add LinkTailPGC

	# search for Break commands
	set idx [lsearch -regexp $cmds {00 .2 .. .. .. .. .. ..}]	;# Break
	while {$idx != -1} {
		set bcmd [lindex $cmds $idx]
		set gcmd [string range $bcmd 0 3]1[string range $bcmd 5 20][format %02X $gotocmd]
		lset cmds $idx $gcmd
		set idx [lsearch -regexp $cmds {00 .2 .. .. .. .. .. ..}]	;# Break
	}
	set_type_cmds pre $pgcvar $cmds

	refresh_current_pgc
	::main::dvd_modified true false
}

proc find_unused_gprms {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	for {set i 0} {$i < 16} {incr i} {
		set used($i) false
	}

	# FP PGC
	foreach i [get_used_gprms ::pgcs(0,0)] {
		set used($i) true
	}

	# VMG Menus
	for {set lu 1} {$lu <= $::menus(0,numlus)} {incr lu} {
		for {set pgc 1} {$pgc <= $::menus(0,$lu,numpgcs)} {incr pgc} {
			foreach i [get_used_gprms ::menus(0,$lu,$pgc)] {
				set used($i) true
			}
		}
	}

	# Titles
	set puos 0
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set ttn 1} {$ttn <= $::pgcs($vts,numpgcs)} {incr ttn} {
			foreach i [get_used_gprms ::pgcs($vts,$ttn)] {
				set used($i) true
			}
		}
	}

	# VTS Menus
	for {set vts 1} {$vts <= $::menus(numvts)} {incr vts} {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				foreach i [get_used_gprms ::menus($vts,$lu,$pgc)] {
					set used($i) true
				}
			}
		}
	}
	
	set txt ""
	for {set i 0} {$i < 16} {incr i} {
		if {! $used($i)} {set txt "$txt$i, "}
	}
	if {$txt==""} {
		set txt "All GPRMs are used!"
	} else {
		set txt "Unused GPRM numbers:\n[string trim $txt ", "]."
	}
	tk_messageBox -title "Find Unused GPRMs" -icon info -message $txt -parent . -type ok
}
proc get_used_gprms {cmds_upvar} {
	upvar $cmds_upvar cmds
	for {set i 0} {$i < 16} {incr i} {
		set used($i) false
	}
	foreach cmd $cmds {
		if {[string range $cmd 0 0] == "*"} continue
		set d [::disasm::disasm $cmd]
		set idx [string first "gprm(" $d]
		while {$idx >= 0} {
			incr idx 5
			scan [string range $d $idx end] "%d" g
			set used($g) true
			set idx [string first "gprm(" $d $idx]
		}
	}
	set allused ""
	for {set i 0} {$i < 16} {incr i} {
		if $used($i) {lappend allused $i}
	}
	return $allused
}

}	;# end of namespace ::utils
